try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

try:
    import pyuda
except ImportError:
    pyuda = None

__all__ = ['OMFITudaValue', 'OMFITudaConnections']

OMFITudaConnections = {}


class OMFITudaValue:
    """
    A wrapper for pyuda calls that behaves like OMFITmdsValue
    """

    def __init__(self, server, shot=None, TDI=None, **kw):
        r"""
        :param server: The device or server from which to get a signal

        :param shot: Which shot

        :param TDI: The signal to fetch

        :param \**kw: Additional keywords that OMFITmdsValue might take, but
            are not necessary for UDA
        """
        if pyuda is None:
            raise RuntimeError('pyuda package is not installed')
        if shot is None:
            raise ValueError('shot must be given')
        if TDI is None:
            raise ValueError('TDI signal must be given')
        if server not in OMFITudaConnections:
            OMFITudaConnections[server] = pyuda.Client()
        self.client = OMFITudaConnections[server]
        self.shot = shot
        self.signal = TDI
        self.udasignal = self.client.get(self.signal, self.shot)

    def data(self):
        return self.udasignal.data[:]

    def dim_of(self, index=0):
        return self.udasignal.dims[index].data

    def units(self):
        return self.udasignal.units

    def units_dim_of(self, index=0):
        return self.udasignal.dims[index].units

    def xarray(self):
        """
        :return: DataArray with information from this node
        """

        data = self.data()
        dims = ['dim_%d' % k for k in range(data.ndim)]
        clist = []
        for k, c in enumerate(dims):
            clist.append(self.dim_of(k))

        if data.shape != tuple([len(k) if np.ndim(k) == 1 else k.shape[ik] for ik, k in enumerate(clist)]):
            dims = dims[::-1]
            clist = clist[::-1]

        coords = {}
        for k, c in enumerate(dims):
            if np.ndim(clist[k]) == 1:
                ck = c
                coords[ck] = ([c], clist[k], {'units': self.units_dim_of(k)})
            else:
                ck = c + '_val'
                coords[ck] = (dims, clist[k], {'units': self.units_dim_of(k)})

        xdata = DataArray(data, dims=dims, coords=coords, attrs={'units': self.units()})
        return xdata



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise


from omfit_classes.omfit_mds import OMFITmdsValue

from matplotlib import pyplot
from matplotlib.widgets import MultiCursor
import numpy as np

__all__ = ['DI_signal', 'DI_DTS_signal', 'DI_file_signal', 'DI_LP_signal', 'DI_GW_signal', 'DI_asdex_signal']


class BadOrMissingDataError(ValueError, doNotReportException):
    """Couldn't find/use data necessary to back one of the signal requests"""


class DI_signal(SortedDict):
    """
    Principal data class used by _Detachment_Indicator.

    Data is fetched as a MDS+ pointname, truncated, ELM filtered, smoothed and remapped to an independent
    variable. See below for inherited objects with specific implementation needs (eg langmuir probes, DTS ...)
    """

    def __init__(self, tag):
        """
        :param tag: string
            a keyword name for this signal. Eg. pointname of MDS+ siganl
        """
        self['tag'] = tag
        self['derived'] = False

    def fetch(self, shotnumber, device):
        """
        Fetch data from MDS+

        :param shotnumber: int
            Shot identifier

        :pram device: string
            Device name, to be used as server name in creation of MDS+ connection
        """

        self['shotnumber'] = shotnumber
        self['device'] = device

        mdsval = OMFITmdsValue(server=device, shot=shotnumber, TDI=self['tag'])
        if not mdsval.check():
            raise BadOrMissingDataError(f"Could not find signal {self['tag']} data for {device}#{shotnumber}")

        self['RAW'] = SortedDict()
        self['RAW']['times'] = mdsval.dim_of(0)  # ms
        self['RAW']['data'] = mdsval.data()
        self['RAW']['units'] = mdsval.units()

    def truncate_times(self, time_range):
        """
        Trim the dataset down by removing values. Helps with speed and to avoid out-of-bounds interpolation errors

        :param time_range: list
            Min and max times for truncation [min, max]
        """

        if time_range is None:
            # Do not truncate
            self['RAW_TRIM'] = SortedDict()
            self['RAW_TRIM']['times'] = self['RAW']['times']
            self['RAW_TRIM']['data'] = self['RAW']['data']
            self['RAW_TRIM']['units'] = self['RAW']['units']
        else:
            tmin_loc = np.argmin(abs(self['RAW']['times'] - time_range[0]))
            tmax_loc = np.argmin(abs(self['RAW']['times'] - time_range[1]))

            self['RAW_TRIM'] = SortedDict()
            self['RAW_TRIM']['times'] = self['RAW']['times'][tmin_loc:tmax_loc]
            self['RAW_TRIM']['data'] = self['RAW']['data'][tmin_loc:tmax_loc]
            self['RAW_TRIM']['units'] = self['RAW']['units']

    def ELM_filter(self, elm_obj, filter_settings):
        """
        Requires that trimmed data already exists.

        :param elm_obj: omfit_elm instance
        A DI_signal doesn't need to know the details of the ELM identification that is identified
        globally for a shot and then shared out to a variety of signals/diagnostics for processing.
        As such, the ELM filtering is done here using a pre-defined omfit_elm object.

        :param filter_settings: Dict
         dict containing the values to be passed to ELM filtering. See OMFITelm for specification
        """

        self['settings_ELM_filter'] = filter_settings  # set a record of what was used
        times_to_filter = self['RAW_TRIM']['times']
        vals = self['RAW_TRIM']['data']
        stimes = np.ones(np.shape(times_to_filter)) * filter_settings['s_time']
        elm_obj.set_filter_params(filter_settings)
        elm_okay = elm_obj(times_to_filter, stimes=stimes)

        self['FILTERED'] = SortedDict()
        self['FILTERED']['times'] = times_to_filter[elm_okay]
        self['FILTERED']['data'] = vals[elm_okay]
        self['FILTERED']['units'] = self['RAW_TRIM']['units']

    def smooth(self, smooth_settings):
        """
        Smooth the filtered data using some kind

        :param smooth_settings" Dict
            A dict of settings as required for the particular kinds of smoothing requested in smooth_settings['smoothtype']
        """
        self['settings_smoothing'] = smooth_settings  # set a record of what was done

        if smooth_settings['smoothtype'] == 'savgol':
            times = self['FILTERED']['times']
            data = scipy.signal.savgol_filter(
                self['FILTERED']['data'], smooth_settings['window_length'], smooth_settings['polyorder'], mode='nearest'
            )

        elif smooth_settings['smoothtype'] == 'median+savgol':
            times = self['FILTERED']['times']
            data_med = scipy.signal.medfilt(self['FILTERED']['data'], smooth_settings['median_length'])
            data = scipy.signal.savgol_filter(data_med, smooth_settings['window_length'], smooth_settings['polyorder'], mode='nearest')

        self['SMOOTHED'] = SortedDict()
        self['SMOOTHED']['times'] = times
        self['SMOOTHED']['data'] = data
        self['SMOOTHED']['units'] = self['FILTERED']['units']

    def remap(self, remap_signal):
        """
        Remap the signal onto an independent variable, which itself is an DI_signal instance
        """
        remap_fun = interp1d(remap_signal['SMOOTHED']['times'], remap_signal['SMOOTHED']['data'], bounds_error=False)
        xdata = remap_fun(self['SMOOTHED']['times'])
        ydata = self['SMOOTHED']['data']

        self['REMAPPED'] = SortedDict()
        self['REMAPPED']['xname'] = remap_signal['tag']
        self['REMAPPED']['xunits'] = remap_signal['SMOOTHED']['units']
        self['REMAPPED']['xdata'] = xdata
        self['REMAPPED']['data'] = ydata
        self['REMAPPED']['units'] = self['SMOOTHED']['units']

    def __call__(self):
        self.plot()

    def plot(self):
        """
        A basic plotting function of the available datatypes.
        """

        pyplot.figure()
        try:
            pyplot.plot(self['RAW']['times'], self['RAW']['data'], color='#87CEFF', label='RAW')
        except Exception:
            print(self['tag'] + ': raw data not available')
        try:
            pyplot.plot(self['RAW_TRIM']['times'], self['RAW_TRIM']['data'], color='#4A708B', label='RAW_TRIM')
        except Exception:
            print(self['tag'] + ': trimmed data not available')
        try:
            pyplot.plot(self['FILTERED']['times'], self['FILTERED']['data'], '+', color='black', label='FILTERED')
        except Exception:
            print(self['tag'] + ': filtered data not available')
        try:
            pyplot.plot(self['SMOOTHED']['times'], self['SMOOTHED']['data'], color='red', label='SMOOTHED')
        except Exception:
            print(self['tag'] + ': smoothed data not available')
        try:
            pyplot.xlim([np.min(self['RAW_TRIM']['times']) * 0.9, np.max(self['RAW_TRIM']['times']) * 1.05])
        except Exception:
            pass
        pyplot.xlabel('Time [ms]')
        pyplot.ylabel(self['RAW']['units'])
        pyplot.title(self['tag'])
        ax = pyplot.gca()
        ax.ticklabel_format(useOffset=False)
        ax.get_xaxis().get_major_formatter().set_scientific(False)
        pyplot.text(1.01, 0.97, self['shotnumber'], rotation=90, transform=ax.transAxes, fontsize=9)
        pyplot.legend(loc='best').draggable(True)
        pyplot.tight_layout()
        pyplot.show(block=False)

    def plot_elm_check(self, xrange=None):

        if xrange is None:
            xrange = [np.min(self['RAW_TRIM']['times']) * 0.9, np.max(self['RAW_TRIM']['times']) * 1.05]

        fig = pyplot.figure()
        ax1 = pyplot.subplot(3, 1, 1)
        ax2 = pyplot.subplot(3, 1, 2, sharex=ax1)

        elm_obj = self['elm_obj']
        elm_obj.elm_detection_plot(
            time_zoom=xrange, crop_data_to_zoom_range=False, show_more_timing=False, show_details=False, fig=fig, axs=[ax1, ax2]
        )
        ax3 = pyplot.subplot(3, 1, 3, sharex=ax1)
        pyplot.plot(self['RAW']['times'], self['RAW']['data'], color='#87CEFF')
        pyplot.plot(self['RAW_TRIM']['times'], self['RAW_TRIM']['data'], color='#4A708B')
        pyplot.plot(self['FILTERED']['times'], self['FILTERED']['data'], '+', color='black')
        pyplot.gcf().multi = MultiCursor(pyplot.gcf().canvas, [ax1, ax2, ax3], color='r', lw=1)
        pyplot.xlabel('Time [ms]')
        pyplot.title(self['tag'])
        pyplot.ylabel(self['RAW']['units'])
        pyplot.tight_layout()


class BadThomsonDataError(BadOrMissingDataError):
    """(Divertor) Thomson Scattering (TS or DTS) data for this shot or channel are unavailable."""


class DI_DTS_signal(DI_signal):
    """
    DTS specific methods. Nothing fancy here, just a parsing of the DTS arrays stored in MDS+. It can't be done
    with DI_signal as the MDS+ arrays are 2D.
    """

    def __init__(self, tag):
        """
        Initialization.

        :param tag: string
        Reference name for quantity. This will be in the form DTS_x_y where x is the quantity (eg DENS, TEMP)
        and y is the channel (eg 0 for the channel closest to the plate). Eg 'DTS_TEMP_0'. Case independent.
        """
        DI_signal.__init__(self, tag)

        _, param, channel = tag.split('_')
        self['tag'] = tag

        if param.lower() == 'dens':
            self['parameter'] = 'ne'
        elif param.lower() == 'temp':
            self['parameter'] = 'te'
        self['chordnumber'] = int(channel)
        self['derived'] = False

    def fetch(self, shotnumber, device):
        """
        Fetch data from MDS+ and split up to find requested quantity on requested channel.

        :param shotnumber: integer
        Shotnumber for data fetch

        :param device:
        Server name for MDS+ data fetch. eg 'DIII-D'
        """
        self['shotnumber'] = shotnumber
        self['device'] = device

        self['RAW'] = SortedDict()

        mdsval = OMFITmdsValue(server=device, shot=shotnumber, TDI="ts" + self['parameter'] + '_div')
        if not mdsval.check():
            raise BadThomsonDataError(f"Could not find any DTS {self['parameter']} data for {device}#{shotnumber}")
        n_dts_ch = len(mdsval.data()[:, 0])
        if self['chordnumber'] >= n_dts_ch:
            raise BadDivThomsonData(
                f"DTS channel index {self['chordnumber']} was requested, "
                f"but the DTS data for {device}#{shotnumber} only cover channels 0-{n_dts_ch}"
            )

        self['RAW']['times'] = mdsval.dim_of(0)  # ms
        self['RAW']['data'] = mdsval.data()[self['chordnumber'], :]

        if self['parameter'] == 'te':
            self['RAW']['data'][self['RAW']['data'] == 0] = np.nan  # indicates bad fit
            self['RAW']['data'][self['RAW']['data'] > 1e3] = np.nan  # not physical and very unhelpful

        if self['parameter'] == 'te':
            self['RAW']['units'] = 'eV'
        elif self['parameter'] == 'ne':
            self['RAW']['units'] = 'm**3'


class DI_file_signal(DI_signal):
    """
    A convenient way to make a DI_signal object when the times and values are coming from file. For example
    a custom-read of some data from an IDL save file. In these cases, it is sometimes easier just to
    pass the raw data straight to the object and then it will be able to handle the filtering, remapping etc. in
    a way that is equivalent to the other diagnostics.
    """

    def __init__(self, tag):
        """
        Initialize DI object with tag as specified.
        :param tag: string
        Tag name for signal, used for identification purposes
        """
        DI_signal.__init__(self, tag)
        self['tag'] = tag

    def fetch(self, shotnumber, times, data, units=''):
        """
        No fetching actually occurs for this subclass. In this case, the otherwise-sourced data is
        simply placed into the attributes that are required by the processing methods in DI_signal.

        :param shotnumber: int
        Shotnumber for reference purposes

        :param times: array-like
        List of times corresponding to data [ms].

        :param data: array-like
        The actual data, 1D array.

        :param units: string
        Description of the units of data.
        """
        self['shotnumber'] = shotnumber
        self['RAW'] = SortedDict()
        self['RAW']['times'] = times  # ms
        self['RAW']['data'] = data
        self['RAW']['units'] = units


class DI_LP_signal(DI_signal):
    """
    Langmuir probe specific methods. Interacts with the Langmuir_Probes module.
    """

    def __init__(self, tag):
        """
        :param tag: string
        reference name for the signal
        """
        DI_signal.__init__(self, tag)

        self['tag'] = tag
        self['derived'] = False

    def fetch(self, shotnumber, probe_tree, param_name, units, processing_settings=None):
        """
        This subclass can't know anything about where the LP data is being saved in the tree. It can however operate
        on one of the probe sub trees. That means the Langmuir_Probe module tree needs to be populated first,
        and then one of the sub trees from its ['OUTPUT'] can be passed into here, alongside the physical quantity
        to be extracted from the tree.

        :param probe_tree: OMFITTree
        a subtree of the Langmuir_Probe module corresponding

        :param param_name: string
        The name of the physical parameter to be extracted from probe tree (eg. JSAT, DENS, or TEMP).

        :param units: OMFITTree
        Descriptors of the units for each of the param_name options. This often lives
        in Langmuir_Toolbox['OUTPUTS']['LP_MDS']['units']

        :param processing_settings: Dict
        settings dict containing filter_settings, smoothing_settings, DOD_tmin, DOD_tmax. See process_data.py
        for full details.
        """

        self['shotnumber'] = shotnumber

        self['RAW'] = SortedDict()
        self['RAW']['times'] = probe_tree['time']
        self['RAW']['data'] = probe_tree[param_name]
        self['RAW']['units'] = units[param_name]


class DI_DOD_signal(DI_signal):
    """
    Derived quantity: Calculate degree of detachment (DOD).
    The process is that jsat for a single probe is calculated as a function of the line averaged density.
    A n**2 fit is then calculated in a time range that should be just the well attached portion of the discharge.
    After the JSAT rollover, JSAT will deviate from the n**2 scaling and the ratio of this scaling to JSAT is
    the degree of detachment. NB: this formulation is just DOD for a single probe. Other methods may be possible
    for some shots. For example, using the integral from a JSAT profile, or a Eich fit to the JSAT profiles.
    """

    def __init__(self, tag):
        """
        Initialization
        :param tag: string
        reference name for signal
        """
        DI_signal.__init__(self, tag)

        self['tag'] = tag
        self['derived'] = True

    def fetch(self, shotnumber, device, probe_tree, units, LP_processing_settings, density_processing_settings):
        """
        This subclass will operate on the probe tree to calculate DOD as per in DI_LP_signal

        :param shotnumber: int
        number of shot of interest

        :param device: string
        server for MDS+ calls
        .
        :param units: OMFITTree
        Descriptors of the units for each of the param_name options. Often empty for DOD

        :param LP_processing_settings: Dict
        settings dict containing filter_settings, smoothing_settings, DOD_tmin, DOD_tmax. See process_data.py
        for full details.

        :param density_processing_settings: Dict
        settings dict containing filter_settings, smoothing_settings for the density processing
        """

        print('Calculating DOD')

        self['shotnumber'] = shotnumber
        self['device'] = device

        DOD_min = LP_processing_settings['DOD_tmin']
        DOD_max = LP_processing_settings['DOD_tmax']
        dod_zeroth_order_term = LP_processing_settings.get('dod_zeroth_order_term', True)
        truncate_range = [LP_processing_settings['truncate_range'][0] - 300, LP_processing_settings['truncate_range'][1] + 300]

        jsat = DI_LP_signal(self['tag'])
        jsat.fetch(shotnumber, probe_tree, 'JSAT', units)
        jsat.truncate_times(truncate_range)
        jsat.ELM_filter(LP_processing_settings['elm_obj'], LP_processing_settings['filter_settings'])
        jsat.smooth(LP_processing_settings['smooth_settings'])

        # The processing for density is handled incrementally to ensure settings align with it being the
        # independent variable for this jsat DI_signal instance
        density_pointname = density_processing_settings.get('pointname', 'density')
        # The den** pointnames include both passes of the laser through the plasma so need 0.5
        # fmt: off
        density_factor = (
            density_processing_settings.get('factor', None)
            or {
                'density': 1e-13,
                'prmtan_neped': 1e-19,
                'denv3': 0.5e-13,
                'denv2': 0.5e-13,
                'denv1': 0.5e-13,
                'denr0': 0.5e-13,
            }.get(density_pointname, 1)
        )
        # fmt: on
        density = DI_signal(density_pointname)
        density.fetch(self['shotnumber'], self['device'])
        density.truncate_times([np.min(jsat['RAW']['times']) - 200, np.max(jsat['RAW']['times']) + 200])
        density.ELM_filter(density_processing_settings['elm_obj'], density_processing_settings['filter_settings'])
        density.smooth(density_processing_settings['smooth_settings'])

        jsat.remap(density)
        xdata = copy.copy(jsat['REMAPPED']['xdata'])
        ydata = copy.copy(jsat['REMAPPED']['data'])
        times = copy.copy(jsat['SMOOTHED']['times'])
        xmin_loc = np.argmin(abs(jsat['SMOOTHED']['times'] - DOD_min))
        xmax_loc = np.argmin(abs(jsat['SMOOTHED']['times'] - DOD_max))

        # This is the real fitting part of JSAT as a function of density. It may seem overkill for a polynomial fit
        # but polyfit was giving me trouble and this will be extensible for more sophisticated fits.
        def DOD_fun(p, xdata):
            """
            Definition of Degree of Detachment. See Loarte NF 38(3): 1998 for details
            :param xdata: arraylike
            Densities
            :return: arraylike
            Degree of Detachment at all densities specified in xdata
            """
            ydata = p[0] * xdata ** 2
            if dod_zeroth_order_term:
                ydata += p[1]
            return ydata

        def fit_fun(p, exp_data=None):
            """
            A function used for optimizing the DOD fit. Compares fit to data and returns least squares difference
            that can be used for minimization purposes

            :param p: arraylike
            Coefficients for fit that are passed to DOD_fun

            :param exp_data: arraylike
            The experimental data to be compared against (eg a single probe's JSAT)

            :return: float
            Least squares difference for this set of p values
            """
            xdata, ydata = exp_data
            fit = DOD_fun(p, xdata)
            lsq_output = np.sum((ydata - fit) ** 2)
            return lsq_output

        fit_fun_partial = functools.partial(fit_fun, exp_data=[xdata[xmin_loc:xmax_loc] * density_factor, ydata[xmin_loc:xmax_loc]])
        p_guess = [1, 0]

        f = scipy.optimize.minimize(fit_fun_partial, p_guess, method='Powell')
        n_scaling = DOD_fun(f.x, xdata * density_factor)
        DOD = n_scaling / ydata
        self['RAW'] = SortedDict()
        self['RAW']['times'] = times
        self['RAW']['data'] = DOD
        self['RAW']['units'] = ''
        self['RAW']['DOD_jsat'] = ydata
        self['RAW']['DOD_density'] = xdata * density_factor
        self['RAW']['DOD_fit'] = n_scaling
        self['RAW']['DOD_time_range'] = [DOD_min, DOD_max]
        if dod_zeroth_order_term:
            self['RAW']['DOD_coefficients'] = f.x
        else:
            self['RAW']['DOD_coefficients'] = [f.x[0], 0]
        self['RAW']['density_pointname'] = density_pointname


class DI_GW_signal(DI_signal):
    """
    A convenience function for calculating the greenwald fraction
    """

    def __init__(self, tag):
        """
        Initialization of greenwald DI_signal
        :param tag: string
        reference name for signal
        """
        DI_signal.__init__(self, tag)

        self['tag'] = tag
        self['derived'] = True

    def fetch(self, shotnumber, device, processing_settings):
        """
        Fetch the mdsplus values for density,aminor, and ip to calculate the Greenwald fraction

        :param shotnumber: integer
        The number of the shot of interest

        :param device: string
        Name of the server to be used for MDS+ calls

        :param processing_settings: dict
        A nested dictionary of settings to be used for smoothing and ELM filtereing. See process_data.py for more
        information
        """

        print('Calculting greenwald fraction')

        self['shotnumber'] = shotnumber
        self['device'] = device

        dataset = {}

        for signal_name in ['aminor', 'ip', 'density']:
            data = DI_signal(signal_name)
            data.fetch(shotnumber, device)
            data.truncate_times(None)
            data.ELM_filter(processing_settings[signal_name]['elm_obj'], processing_settings[signal_name]['filter_settings'])
            data.smooth(processing_settings[signal_name]['smooth_settings'])
            data.smoothed_interp_fun = interp1d(data['SMOOTHED']['times'], data['SMOOTHED']['data'])
            dataset[signal_name] = data

        # Determine the minimum range of density times for which all signals are valid to provide a time-base that can
        # be reliably interpolated onto.
        t_mins = []
        t_maxes = []
        for signal_name in dataset:
            t_mins.append(np.min(dataset[signal_name]['SMOOTHED']['times']))
            t_maxes.append(np.max(dataset[signal_name]['SMOOTHED']['times']))
        dens_tmin_loc = np.argmin(abs(dataset['density']['SMOOTHED']['times'] - np.max(t_mins))) + 1
        dens_tmax_loc = np.argmin(abs(dataset['density']['SMOOTHED']['times'] - np.min(t_maxes))) - 1

        times = dataset['density']['SMOOTHED']['times'][dens_tmin_loc:dens_tmax_loc]
        aminor = dataset['aminor'].smoothed_interp_fun(times)
        ip = dataset['ip'].smoothed_interp_fun(times)
        density = dataset['density'].smoothed_interp_fun(times)

        greenwald_fraction = density * 3.14e-8 * aminor * aminor / ip

        self['RAW'] = SortedDict()
        self['RAW']['times'] = times
        self['RAW']['data'] = greenwald_fraction
        self['RAW']['units'] = ''


class DI_asdex_signal(DI_signal):
    """
    A convenience function for calculating the various kinds of neutral compression from the gauges
    """

    def __init__(self, tag):
        """
        Initialization of greenwald DI_signal
        :param tag: string
        reference name for signal
        """
        DI_signal.__init__(self, tag)

        self['tag'] = tag
        self['derived'] = True

    def fetch(self, shotnumber, device, processing_settings):
        """
        Fetch the mdsplus values for calculation of the neutral compression in the SAS.

        :param shotnumber: integer
        The number of the shot of interest

        :param device: string
        Name of the server to be used for MDS+ calls

        :param processing_settings: dict
        A nested dictionary of settings to be used for smoothing and ELM filtereing. See process_data.py for more
        information
        """

        print('Calculting ASDEX gauge SAS compression')

        self['shotnumber'] = shotnumber
        self['device'] = device
        dataset = {}

        for signal_name in ['asdex_sas1a', 'asdex_sas1b']:
            data = DI_signal(signal_name)
            data.fetch(shotnumber, device)
            data.truncate_times([-500, np.max(data['RAW']['times'])])  # Remove extreme values before shot begins
            dataset[signal_name] = data

        # ASDEX gagues are on the same time base, so they can just be divided

        self['RAW'] = SortedDict()
        self['RAW']['times'] = dataset['asdex_sas1a']['RAW_TRIM']['times']
        self['RAW']['data'] = dataset['asdex_sas1a']['RAW_TRIM']['data'] / dataset['asdex_sas1b']['RAW_TRIM']['data']
        self['RAW']['units'] = ''



print('External imports...')

from omfit_classes import unix_os as os
import sys

# the OMFITsrc variable stores the directory where OMFIT is running from
OMFITsrc = os.path.abspath(os.path.dirname(__file__) + os.sep + '..')

# load base utilities
from omfit_classes.utils_base import *

# this defines the command line that is shown when OMFIT is used at terminal
# Note that setting ps1 is also required to avoid crashing with Matplotlib v1.4 (matplotlib bug)
sys.ps1 = 'OMFIT >>> '

# monkey patching of np `array`, `asarray` and `asanyarray` to fix array behavior with OMFIT dynamic expressions
from numpy.core.multiarray import array as _array


def array(object, dtype=None, *, copy=True, order='K', subok=False, ndmin=0, like=None):
    object = evalExpr(object)
    if like is not None:
        return _array(object, dtype=dtype, copy=copy, order=order, subok=subok, ndmin=ndmin, like=None)
    else:
        return _array(object, dtype=dtype, copy=copy, order=order, subok=subok, ndmin=ndmin)


array.__doc__ = _array.__doc__

from numpy.core.fromnumeric import asarray as _asarray


def asarray(a, dtype=None, order=None, *, like=None):
    a = evalExpr(a)
    if like is not None:
        return _asarray(a, dtype=dtype, order=order, like=like)
    else:
        try:
            return _asarray(a, dtype=dtype, order=order)
        except Exception:
            print("=!=!=!=")
            print(a)
            print(type(a))
            print("=!=!=!=")
            raise


asarray.__doc__ = _asarray.__doc__

from numpy.core.fromnumeric import asanyarray as _asanyarray


def asanyarray(a, dtype=None, order=None):
    a = evalExpr(a)
    return _asanyarray(a, dtype=dtype, order=order)


asanyarray.__doc__ = _asanyarray.__doc__

from numpy.core import multiarray as _multiarray

_multiarray.array = array
_multiarray.asarray = asarray
_multiarray.asanyarray = asanyarray

from numpy.core import fromnumeric as _fromnumeric

_fromnumeric.array = array
_fromnumeric.asarray = asarray
_fromnumeric.asanyarray = asanyarray

from numpy.core import shape_base as _shape_base

_shape_base.array = array
_shape_base.asarray = asarray
_shape_base.asanyarray = asanyarray

# Monkey patch gradient so it doesn't complain about 0-d arrays as not being scalars
import numpy.lib.function_base
from numpy.lib.function_base import gradient as _gradient


def gradient(f, *args, **kwargs):
    return _gradient(f, *[x * 1.0 for x in args], **kwargs)


gradient.__doc__ = _gradient.__doc__
numpy.lib.function_base._gradient = _gradient
numpy.lib.function_base.gradient = gradient

import numpy as np
from numpy.lib.recfunctions import *
import bisect

warnings.simplefilter('ignore', np.RankWarning)
np.array = array
np.asarray = asarray
np.asanyarray = asanyarray
np.gradient = gradient

# matplotlib (if matplotlib module was already loaded, keep the same rc parameters)
if 'matplotlib' in sys.modules:
    if 'matplotlib.pyplot' in sys.modules:
        raise ImportError('Because OMFIT monkey patches matplotlib figure and ' 'axes, you must import OMFIT before pyplot')
    _rc_updates = copy.deepcopy(sys.modules['matplotlib'].rcParams)
    import matplotlib

else:
    import matplotlib

    _rc_updates = {
        'figure.facecolor': 'white',
        'axes.formatter.use_mathtext': True,
        'legend.labelspacing': 0.2,
        'legend.fontsize': 'medium',
        'savefig.format': 'pdf',
        'pdf.fonttype': 42,
        'ps.useafm': True,
        'axes.formatter.limits': [-3, 3],
        'axes.formatter.useoffset': False,
        'keymap.fullscreen': 'f',
        'keymap.home': 'h',
        'keymap.back': '',
        'keymap.forward': '',
        'keymap.pan': 'p',
        'keymap.zoom': 'z',
        'keymap.save': 's',
        'keymap.grid': 'g',
        'keymap.yscale': 'l',
        'keymap.xscale': 'L',
        'mathtext.default': 'regular',
        'savefig.transparent': True,
    }

if hasattr(matplotlib, 'get_data_path'):
    _mpl_font_path = os.sep.join([matplotlib.get_data_path(), 'fonts', 'ttf'])
else:
    _mpl_font_path = os.sep.join([matplotlib.rcParams['datapath'], 'fonts', 'ttf'])
if not os.path.exists(os.sep.join([_mpl_font_path, 'Muli.ttf'])) and os.access(_mpl_font_path, os.W_OK):
    _font_path = os.sep.join([OMFITsrc, 'extras', 'graphics', 'fonts'])
    for _f in os.listdir(_font_path):
        _mpl_fname = os.sep.join([_mpl_font_path, _f])
        if not os.path.exists(_mpl_fname):
            os.system('cp %s %s' % (os.sep.join([_font_path, _f]), _mpl_fname))

# remove fontCache file if produced before Muli got installed (previous lines added)
import matplotlib.font_manager

_fontcachefn = getattr(matplotlib.font_manager, '_fmcache', '')
if os.path.exists(_fontcachefn) and os.path.getmtime(_fontcachefn) < 1586983275:
    print('Removing font cache to get updated fonts')
    os.unlink(_fontcachefn)
if os.environ.get('OMFIT_NO_GUI', '0') == '1' or not os.environ.get("DISPLAY"):
    matplotlib.use('agg')
else:
    matplotlib.use('TkAgg')
# ===========================
import matplotlib.axes


def _imag_arg(arg):
    if np.any(np.iscomplex(arg)):
        return np.imag(arg)
    else:
        return arg


def _real_arg(arg):
    if np.any(np.iscomplex(arg)):
        return np.real(arg)
    else:
        return arg


class dummyLegend(object):
    def draggable(*args, **kw):
        pass

    def set_draggable(*args, **kw):
        pass


# Avoid red text in console
import logging

with quiet_environment():
    logging.getLogger('matplotlib.legend').warning('')

_orig_matplotlib_axes_Axes = matplotlib.axes.Axes


class Axes(_orig_matplotlib_axes_Axes):
    def plot(self, *args, **kw):
        linestyles = list(matplotlib.lines.lineStyles.keys())
        newargs = list(map(_real_arg, args))
        lines1 = _orig_matplotlib_axes_Axes.plot(self, *newargs, **kw)
        lines2 = []
        if np.any([np.any(np.iscomplex(arg).flat) for arg in args]):
            newargs = list(map(_imag_arg, args))
            lines2 = _orig_matplotlib_axes_Axes.plot(self, *newargs, **kw)
            for l1, l2 in zip(lines1, lines2):
                l2.set_color(l1.get_color())
                ils = linestyles.index(l1.get_linestyle())
                l2.set_linestyle(linestyles[ils + 1])
                l2.set_label(r'$\Im$ ' + l1.get_label())
                l1.set_label(r'$\Re$ ' + l1.get_label())
            pyplot.gca().legend(ncol=max(1, len(self.lines) // 6))
        self.callbacks.connect('xlim_changed', self.axes_downsample)
        return lines1 + lines2

    def legend(self, *args, **kw):
        r"""
        Modified matplotlib legend method.

        :param handles: list of artists
            Artists (lines, patches) to be added to the legend

        :param labels: list of strings
            Label to use in the legend

        :param no_duplicates: bool
            Keeps only one of each unique legend label and deletes the others

        :param alpha_reset: float (optional)
            If not None, sets transparency of legend items to specific value.
            alpha_reset=1.0 is useful if there are many transparent symbols
            but you want one full opacity symbol to represent the group.

        :param extra_note: string
            Add text to the bottom of the legend with no marker

        :param text_same_color: bool
            Change the color of the text to match the plot object.

        :param hide_markers: bool
            Hide legend markers or lines and just show legend text.

        :param draggable: bool
            Make the legend draggable.

        :param \**kw: keyword arguments passed to original matplotlib legend method

        :return: legend object

        Original legend docstring:
        """
        kwargs = kw
        no_duplicates = kw.pop('no_duplicates', False)
        alpha_reset = kw.pop('alpha_reset', None)
        extra_note = kw.pop('extra_note', None)
        text_same_color = kw.pop('text_same_color', False)
        hide_markers = kw.pop('hide_markers', False)
        draggable = kw.pop('draggable', True)

        import matplotlib.legend as mleg

        logging.getLogger("matplotlib").setLevel(logging.CRITICAL)  # Prevent sub-critical warnings from displaying
        if hasattr(mleg, '_parse_legend_args'):
            handles, labels, extra_args, kwargs = mleg._parse_legend_args([self], *args)
            if not handles:
                return dummyLegend()
        else:
            # Begin copy from matplotlib.axes._axes.legend
            handlers = kwargs.get('handler_map', {}) or {}

            # Support handles and labels being passed as keywords.
            handles = kwargs.pop('handles', None)
            labels = kwargs.pop('labels', None)

            if (handles is not None or labels is not None) and len(args):
                warnings.warn("You have mixed positional and keyword arguments; some input will be discarded.")

            # If got both handles and labels as kwargs, make same length
            if handles and labels:
                handles, labels = list(zip(*list(zip(handles, labels))))

            elif handles is not None and labels is None:
                labels = [handle.get_label() for handle in handles]
                for label, handle in zip(labels[:], handles[:]):
                    if label.startswith('_'):
                        warnings.warn(
                            'The handle {!r} has a label of {!r} '
                            'which cannot be automatically added to the legend.'.format(handle, label)
                        )
                        labels.remove(label)
                        handles.remove(handle)

            elif labels is not None and handles is None:
                # Get as many handles as there are labels.
                handles = [handle for handle, label in zip(self._get_legend_handles(handlers), labels)]

            # No arguments - automatically detect labels and handles.
            elif len(args) == 0:
                handles, labels = self.get_legend_handles_labels(handlers)
                if not handles:
                    return dummyLegend()

            # One argument. User defined labels - automatic handle detection.
            elif len(args) == 1:
                (labels,) = args
                # Get as many handles as there are labels.
                handles = [handle for handle, label in zip(self._get_legend_handles(handlers), labels)]

            # Two arguments:
            #   * user defined handles and labels
            elif len(args) == 2:
                handles, labels = args

            else:
                raise TypeError('Invalid arguments to legend.')
            # End copy from matplotlib.axes._axes.legend

        # Modify list of handles and labels in legend to prevent duplicate entries
        if no_duplicates:
            i = np.arange(len(labels))
            legend_idxs = np.array([])
            unique_labels = tolist(set(labels))
            for ul in unique_labels:
                legend_idxs = np.append(legend_idxs, [i[np.array(labels) == ul][0]])
            handles = [handles[int(idx)] for idx in legend_idxs]
            labels = [labels[int(idx)] for idx in legend_idxs]

        # Add extra note
        if extra_note is not None:
            for note in tolist(extra_note):
                handles += [matplotlib.patches.Patch(alpha=0)]
                labels += [note]

        # Reset alpha (opacity) value of legend markers to a specified value
        if alpha_reset is not None:
            # Because deepcopy doesn't work on legend handles, we have to:
            # 1. Go through all the handles on the plot, backup the original alpha value, and set the new one
            # 2. Draw the legend
            # 3. Restore all the alpha values on the plot to their original value
            alphass = [None] * len(handles)  # Make a list to hold lists of saved alpha values
            for k, handle in enumerate(handles):  # Loop through the legend entries
                try:  # If handle was a simple list of parts, then this will work
                    alphass[k] = handle.get_alpha()  # This probably only works for a plot with a few simple line plots
                    handle.set_alpha(alpha_reset)
                except AttributeError:
                    alphas = [None] * len(handle)  # Make a list to hold the alphas of the pieces of this legend entry
                    for i, h in enumerate(handle):
                        # Loop through the pieces of this legend entry (there could be a line and a marker, for example)
                        try:  # If handle was a simple list of parts, then this will work
                            alphas[i] = h.get_alpha()
                            h.set_alpha(alpha_reset)
                        except AttributeError:
                            # If handle was a list of parts which themselves were made up of smaller sub-components,
                            # then we must go one level deeper still. This was needed for the output of errorbar() and
                            # may not be needed for simpler plot objects
                            alph = [None] * len(h)
                            for j, hh in enumerate(h):
                                alph[j] = hh.get_alpha()
                                hh.set_alpha(alpha_reset)
                            alphas[i] = alph  # Save the list of alpha values for the sub-components of this piece of
                            #                   this legend entry
                    alphass[k] = alphas  # Save the list of alpha values for the pieces of this legend entry

        if len(handles):
            kw['labels'] = labels
            if hide_markers:
                # Make invisible dummy handles to replace the real legend handles
                kw['handles'] = [matplotlib.patches.Patch(alpha=0)] * len(handles)
                kw['handlelength'] = 0
                kw['handletextpad'] = 0
            else:
                kw['handles'] = handles
            the_legend = _orig_matplotlib_axes_Axes.legend(self, **kw)
        else:
            the_legend = None

        if the_legend is None:
            # if plot is empty, legend() returns `None`
            # This hack prevents raising an error if one does:
            # >> legend().draggable(True)
            printd('legend monkey match returning dummyLegend')
            return dummyLegend()

        if hide_markers:
            for item in the_legend.legendHandles:
                item.set_visible(False)

        if text_same_color:
            for handle, leg_text in zip(handles, the_legend.get_texts()):
                h0 = tolist(handle)[0]
                if hasattr(h0, 'get_color'):
                    handle_color = h0.get_color()  # Works for lines, errorbars, ubands, etc.
                elif hasattr(h0, 'get_facecolor'):
                    handle_color = h0.get_facecolor()  # Works for scatter
                elif hasattr(h0, 'get_edgecolor'):
                    handle_color = h0.get_edgecolor()
                else:
                    handle_color = None  # Give up and leave it alone.
                if handle_color is not None:
                    # Handle color needs to be sanitized because some objects return things like [[1, 0, 0, 1]]
                    if isinstance(handle_color, str):
                        pass  # Already good
                    elif isinstance(handle_color[0], str):
                        handle_color = handle_color[0]
                    else:
                        handle_color = np.squeeze(handle_color)
                    leg_text.set_color(handle_color)

        if alpha_reset is not None:
            # Now we have to restore the original alpha values
            for k, handle in enumerate(handles):  # Loop through legend items to restore original alphas on the plot
                try:
                    handle.set_alpha(alphass[k])
                except AttributeError:
                    for i, h in enumerate(handle):  # Loop through pieces of this legend item to restore alpha values
                        try:
                            h.set_alpha(alphass[k][i])
                        except AttributeError:
                            for j, hh in enumerate(h):
                                # Loop through sub-components of this piece of this legend item to restore alpha values
                                hh.set_alpha(alphass[k][i][j])

        if draggable and the_legend:
            try:
                if hasattr(the_legend, 'set_draggable'):
                    the_legend.set_draggable(True)
                else:
                    the_legend.draggable(True)
            except Exception:
                pass

        return the_legend

    legend.__doc__ += _orig_matplotlib_axes_Axes.legend.__doc__

    def set_downsampling(self, npts):
        """
        Set the minimum number of points before downsampling

        :param npts: If plotting more than npts points in a given curve, then downsample the line
        """
        self.npts = npts

    def axes_downsample(self, ax):
        """
        Down-sample original data in each Line2D object in lines
        (either x or y must be numerical and monotonically increasing)
        """
        npts = getattr(self, 'npts', 1e4)
        for line in ax.lines:
            if not isinstance(line.get_xdata(), np.ndarray):
                continue  # odd bug with xdata just a float in uerrorbar
            if ((not hasattr(line, '_xinit') and len(line.get_xdata()) > npts)) or (hasattr(line, '_xinit') and line._xinit is not None):
                # do not downsample arrays that are strings, boolean, or object type
                for check in [np.asarray(line.get_xdata()).dtype.name, np.asarray(line.get_ydata()).dtype.name]:
                    if 'string' in check or check == 'object' or check == 'bool':
                        line._xinit = None
                        return
                if not hasattr(line, '_xinit'):
                    tmpx = np.sign(np.diff(line.get_xdata()))
                    tmpy = np.sign(np.diff(line.get_ydata()))
                    if np.all(tmpx == tmpx[0]) or np.all(tmpy == tmpy[0]):
                        line.downsample(npts)
                else:
                    line.downsample(npts)

    def format_xydata(self, xy, theaxis):
        """
        Override default matplotlib data formatter to make fixed-width cursor coordinates in the toolbar

        Changing the length of the cursor's message box causes the size of the figure window to change,
        which is distracting. The figure itself may try to stay centered in the window, so it might jump
        around, or it might change size as the cursor moves.

        Formats using "g" give a consistent number of sig-figs, not a consistent string length, hence the
        if/else business with "f" and "e" format codes.
        """

        isdate = isinstance(theaxis.get_major_formatter(), matplotlib.dates.AutoDateFormatter)
        if isdate:
            # return (default_formatter or theaxis.get_major_formatter().format_data_short)(xy)
            return matplotlib.dates.DateFormatter('%Y-%m-%d %H:%M:%S')(xy)
        elif (abs(xy) >= 1e5) or ((xy != 0) and (abs(xy) <= 1e-3)):
            return f'{xy:+12.5e}'
        else:
            return f'{xy:+12.5f}'

    def format_xdata(self, x):
        return self.format_xydata(x, self.xaxis)

    def format_ydata(self, y):
        return self.format_xydata(y, self.yaxis)

    def format_coord(self, x, y):
        """
        Modified message in toolbar to alert users when a plot is interactive.
        """
        prefix = ''
        if hasattr(self, 'is_interactive') and self.is_interactive:
            prefix = 'Interactive, '
        return prefix + _orig_matplotlib_axes_Axes.format_coord(self, x, y)


_orig_matplotlib_legend_Legend = matplotlib.legend.Legend
if hasattr(_orig_matplotlib_legend_Legend, 'set_draggable'):
    orig_draggable = _orig_matplotlib_legend_Legend.set_draggable
else:
    orig_draggable = _orig_matplotlib_legend_Legend.draggable


def draggable(self, state=True, use_blit=False, update="loc"):
    # This monkey patch changes the default value of state from None (toggle) to True (turn on draggability).
    return orig_draggable(self, state=state, use_blit=use_blit, update=update)


_orig_matplotlib_legend_Legend.draggable = draggable
matplotlib.axes.Axes = Axes


# ===========================
from matplotlib import pyplot


def use_subplot(fig, *args, **kw):
    """
    Same as add_subplot, but if the figure already has a subplot
    with key (*args*, *kwargs*) then it will simply make that subplot
    current and return it.
    """
    pyplot.figure(num=fig.number)
    return pyplot.subplot(*args, **kw)


matplotlib.figure.Figure.use_subplot = use_subplot

matplotlib.style.use('classic')
matplotlib.rcParams.update(_rc_updates)
matplotlib.rcParams['interactive'] = True
matplotlib.rcParams['tk.window_focus'] = True

import colorsys

from numpy import nan, NaN
import pylab
import scipy

try:
    import dask.array  # Need to do this before monkeypatching np
except ImportError:
    pass
from scipy.interpolate import *

# patch interp1d so that it can handle a single data point
_interp1d = interp1d


class interp1d(_interp1d):
    __doc__ = _interp1d.__doc__

    def __init__(self, x, y, kind='linear', axis=-1, copy=True, bounds_error=None, fill_value=np.nan, assume_sorted=False, **kw):
        if np.atleast_1d(x).size == 1:
            x = np.atleast_1d(x)
            x = np.array([x[0] - np.finfo(float).epsneg, x[0], x[0] + np.finfo(float).eps])
            y = np.atleast_1d(y)
            y = np.array([y[0], y[0], y[0]])
        _interp1d.__init__(
            self, x, y, kind=kind, axis=axis, copy=copy, bounds_error=bounds_error, fill_value=fill_value, assume_sorted=assume_sorted, **kw
        )


interpolate.interp1d = interp1d
scipy.interpolate.interp1d = interp1d

from scipy import interpolate
from scipy import integrate
from scipy import optimize
from scipy import ndimage
import scipy.io.netcdf as netcdf
from scipy import constants

# --------------
import uncertainties
import uncertainties.unumpy as unumpy
from uncertainties.unumpy import nominal_values, std_devs, uarray
from uncertainties import ufloat, ufloat_fromstr

_isnan = np.isnan


def isnan(x, *args):
    try:
        return _isnan(x, *args)
    except TypeError:
        return unumpy.isnan(x, *args)


isnan.__doc__ = _isnan.__doc__
np.isnan = isnan
pylab.isnan = isnan
scipy.isnan = isnan

_isinf = np.isinf


def isinf(x, *args):
    try:
        return _isinf(x, *args)
    except TypeError:
        return unumpy.isinf(x, *args)


isinf.__doc__ = _isinf.__doc__
np.isinf = isinf
pylab.isinf = isinf
scipy.isinf = isinf

_isfinite = np.isfinite


def isfinite(x, *args):
    try:
        return _isfinite(x, *args)
    except TypeError:
        if np.iterable(x):
            return np.reshape(list(map(bool, 1 - (isinf(x.flat) | np.isnan(x.flat)))), x.shape)
        else:
            return bool(1 - (isinf(x) | np.isnan(x)))


isfinite.__doc__ = _isfinite.__doc__
np.isfinite = isfinite
pylab.isfinite = isfinite
scipy.isfinite = isfinite

# ------------------------------------
# data
if os.environ['USER'] in ['eldond']:
    warnings.filterwarnings('ignore', category=ImportWarning)
    warnings.filterwarnings('always', category=UserWarning, message='.*ModuleNotFoundError.*')
import netCDF4
import csv
import fortranformat

import zipfile
import collections

# strings
import re
import fnmatch
import string
import textwrap
import ast
import difflib

# exec
import subprocess
import multiprocessing

# system
import platform
import signal
import atexit
import shutil
import glob
import struct
import stat
import filecmp

# python
import weakref
import types
import copy
import itertools
import functools
import gc
import importlib
import unittest

# remote
import socket

# time
import datetime
import dateutil
import time

# buffers
import ctypes

import io

# introspection
import inspect
import traceback
import pydoc

# web
import requests
import json

if os.name == 'nt':
    try:
        import wexpect as pexpect
    except Exception as _excp:
        pexpect = None
        warnings.warn('No `wexepect` support: ' + repr(_excp))
else:
    try:
        import pexpect
    except Exception as _excp:
        pexpect = None
        warnings.warn('No `pexepect` support: ' + repr(_excp))

try:
    import pidly
except Exception as _excp:
    pidly = None
    warnings.warn('No `pidly` support: ' + repr(_excp))

try:
    with warnings.catch_warnings(record=True) as w:
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        import boto3
        from boto3.dynamodb.conditions import Attr
        import decimal
except Exception as _excp:
    boto3 = None
    warnings.warn('No `boto3` support: ' + repr(_excp))

try:
    import pandas

    warnings.filterwarnings('ignore', message="Pandas doesn't allow columns to be created via a new attribute name*")
except Exception as _excp:
    pandas = None
    warnings.warn('No `pandas` support! DataArray and Dataset not will not be available : ' + repr(_excp))

try:
    with warnings.catch_warnings(record=True) as w:
        warnings.filterwarnings('ignore', category=DeprecationWarning)
        warnings.filterwarnings('ignore', category=FutureWarning)
        import xarray
except Exception as _excp:
    try:
        import xray as xarray

        warnings.warn("Package `xray` is deprecated. Please upgrade to `xarray`")
    except Exception:
        xarray = None
        warnings.warn('No `xarray` support! DataArray and Dataset not will not be available : ' + repr(_excp))

try:
    import lmfit
except Exception as _excp:
    lmfit = None
    warnings.warn('No `lmfit` support: ' + repr(_excp))

try:
    import gpr1dfusion
except Exception as _excp:
    gpr1dfusion = None
    warnings.warn('No `gpr1dfusion` support! Corresponding Gaussian Process functions will not be available : ' + repr(_excp))
try:
    import jedi
except Exception as _excp:
    jedi = None
    warnings.warn('No jedi support')
else:
    try:
        jedi.preload_module('numpy', 'pylab')
    except Exception:
        # purge jedi cache directory if failing to load cache files
        if os.path.exists(jedi.settings.cache_directory):
            shutil.rmtree(jedi.settings.cache_directory, ignore_errors=True)
        jedi.preload_module('numpy', 'pylab')

from omfit_classes.utils_base import summarize_installed_packages

print('Sanity check Python packages required by OMFIT')
_rc, _ = summarize_installed_packages(required=True, optional=False, verbose=False)
if _rc:
    summarize_installed_packages(required=True, optional=False, verbose=True)
    print(
        '''
Some Python package required by OMFIT is missing. You could:
1. Update the python environment following these instructions: https://omfit.io/install.html
2. Run an older version of OMFIT that does not have this requirement
'''
    )
    exit(_rc)



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

if 'CDB_PATH' in os.environ:
    _pycdb_pythonpath = os.path.abspath(os.environ['CDB_PATH'])
    if _pycdb_pythonpath not in [os.path.abspath(_path) for _path in sys.path]:
        sys.path.insert(1, _pycdb_pythonpath)
        print('* Path to CDB has been added: ' + _pycdb_pythonpath)
    else:
        print('* Path to CDB was found' + _pycdb_pythonpath)
    if 'PYTHONPATH' in os.environ and os.environ['PYTHONPATH']:
        if _pycdb_pythonpath not in os.environ['PYTHONPATH']:
            os.environ['PYTHONPATH'] = _pycdb_pythonpath + ':' + os.environ['PYTHONPATH']
    else:
        os.environ['PYTHONPATH'] = _pycdb_pythonpath
try:
    import pyCDB
    from pyCDB.client import CDBClient
except Exception as _excp:
    pyCDB = None
    CDBClient = None

__all__ = ['OMFITcdb', 'OMFITcdbValue', 'pyCDB']


class OMFITcdbValue(object):
    def __init__(self, **kw):
        self.alias = kw.pop('alias', None)
        if 'shot' in kw:
            kw['record_number'] = kw.pop('shot')
        self.user_kw = kw
        self._kw = kw
        self._data = None

    def __getattr__(self, attr):
        attr_list = [
            'axes',
            'daq_attachment',
            'daq_parameters',
            'data',
            'description',
            'file_ref',
            'get_axes_info',
            'get_log_level',
            'get_ndim',
            'gs_parameters',
            'gs_ref',
            'info',
            'log_level',
            'name',
            'ndim',
            'plot',
            'ref',
            'set_log_level',
            'set_ndim',
            'time_axis',
            'units',
            'url',
        ]
        if attr in attr_list + ['axis%d' % k for k in range(1, 7)]:
            self.fetch()
        return getattr(self._data, attr)

    def fetch(self):
        if self._data is None:
            if 'CDBclient' not in OMFITaux:
                OMFITaux['CDBclient'] = CDBClient()
            client = OMFITaux['CDBclient']
            if 'CDB_cache' not in OMFIT['scratch']:
                OMFIT['scratch']['CDB_cache'] = {}
            if tuple(self.user_kw.values()) not in OMFIT['scratch']['CDB_cache']:
                OMFIT['scratch']['CDB_cache'][tuple(self.user_kw.values())] = client.get_signal(**self.user_kw)
            self._data = OMFIT['scratch']['CDB_cache'][tuple(self.user_kw.values())]
            self._kw = self._data.__dict__

    @property
    def _description(self):
        if 'description' in self._kw and self._kw['description']:
            return self._kw['description']
        elif 'generic_signal_name' in self._kw and self._kw['generic_signal_name']:
            return re.sub('_', ' ', self._kw['generic_signal_name'])
        elif 'generic_signal_id' in self._kw and self._kw['generic_signal_id']:
            return 'signal id: %d' % self._kw['generic_signal_id']

    def plot(self):
        self.fetch()
        tmp = self._data.plot(fig=pyplot.gcf())
        pyplot.gcf().suptitle(self._description)
        return tmp

    def __str__(self):
        description = self._description
        tmp = [description, '=' * 21]
        tmp.append('%s: %s' % ('alias'.ljust(20), self.alias))
        for k in self._kw:
            if k in ['ref', 'data', 'axes']:
                continue
            if isinstance(self._kw[k], pyCDB.client.CDBSignal):
                tmp.append('%s: %s' % (k.ljust(20), 'CDB signal --> ' + str(self._kw[k].name)))
            elif isinstance(self._kw[k], pyCDB.pyCDBBase.OrderedDict):
                continue
            else:
                tmp.append('%s: %s' % (k.ljust(20), self._kw[k]))
        return '\n'.join(tmp)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, repr(self._kw)[1:-1])

    @property
    def help(self):
        units = ''
        if 'units' in self._kw and self._kw['units']:
            units = '[%s] ' % self._kw['units']
        return units + self._description

    def __tree_repr__(self):
        if self._data is not None:
            values = [self.data, ['MDSactive']]
        else:
            values = [self.help, []]
        return values


class OMFITcdb(SortedDict):
    def __init__(self, shot):
        SortedDict.__init__(self, sorted=True)
        self.shot = shot
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        if 'CDBclient' not in OMFITaux:
            OMFITaux['CDBclient'] = CDBClient()
        client = OMFITaux['CDBclient']
        tmp = client.get_generic_signal_references(record_number=self.shot)
        groups = {k.data_source_id: k for k in client.get_data_source_references()}
        for k in tmp:
            group_name = groups[k.data_source_id].name
            self.setdefault(group_name, {})
            self[group_name][k.generic_signal_name] = OMFITcdbValue(record_number=self.shot, **k)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, 'shot=%d' % self.shot)

    def __tree_repr__(self):
        return '#%d @ COMPASS' % self.shot, []


class OMFITcdb(SortedDict):
    def __init__(self, shot):
        SortedDict.__init__(self, sorted=True)
        self.shot = shot
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        if 'CDBclient' not in OMFITaux:
            OMFITaux['CDBclient'] = CDBClient()
        client = OMFITaux['CDBclient']
        tmp = client.get_generic_signal_references(record_number=self.shot)
        groups = {k.data_source_id: k for k in client.get_data_source_references()}
        for k in tmp:
            group_name = groups[k.data_source_id].name
            self.setdefault(group_name, {})
            self[group_name][k.generic_signal_name] = OMFITcdbValue(record_number=self.shot, **k)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, 'shot=%d' % self.shot)

    def __tree_repr__(self):
        return '#%d @ COMPASS' % self.shot, []



"""
Functions for adding DIII-D hardware description data to the IMAS schema by writing data to ODS instances
"""

try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

import numpy as np
import copy
from omfit_classes.omfit_ascii import OMFITascii
from omfit_classes.omfit_mds import OMFITmds, OMFITmdsValue
from omfit_classes.utils_base import compare_version
from omfit_classes.omfit_omas_utils import ensure_consistent_experiment_info

try:
    import omas

    imas_version = omas.ODS().imas_version
except Exception:
    imas_version = '0.0'

__all__ = ['OMFITd3dcompfile']


# Decorators
def _id(obj):
    """Trivial decorator as an alternative to make_available()"""
    return obj


def make_available(f):
    """Decorator for listing a function in __all__ so it will be readily available in other scripts"""
    __all__.append(f.__name__)
    return f


def make_available_if(condition):
    """
    Make available only if a condition is met

    :param condition: bool
        True: make_available decorator is used and function is added to __all__ to be readily available
        False: _id decorator is used, function is not added to __all__, and it cannot be accessed quite so easily
            Some functions inspect __all__ to determine which hardware systems are available
    """
    if condition:
        return make_available
    return _id


# Utilities
def printq(*args):
    return printd(*args, topic='omfit_omas_d3d')


# Hardware descriptor functions
@make_available
def setup_gas_injection_hardware_description_d3d(ods, shot):
    """
    Sets up DIII-D gas injector data.

    R and Z are from the tips of the arrows in puff_loc.pro; phi from angle listed in labels in puff_loc.pro .
    I recorded the directions of the arrows on the EFITviewer overlay, but I don't know how to include them in IMAS, so
    I commented them out.

    Warning: changes to gas injector configuration with time are not yet included. This is just the best picture I could
    make of the 2018 configuration.

    Data sources:
    EFITVIEWER: iris:/fusion/usc/src/idl/efitview/diagnoses/DIII-D/puff_loc.pro accessed 2018 June 05, revised 20090317
    DIII-D webpage: https://diii-d.gat.com/diii-d/Gas_Schematic accessed 2018 June 05
    DIII-D wegpage: https://diii-d.gat.com/diii-d/Gas_PuffLocations accessed 2018 June 05

    Updated 2018 June 05 by David Eldon

    :return: dict
        Information or instructions for follow up in central hardware description setup
    """
    if shot < 100775:
        warnings.warn('DIII-D Gas valve locations not applicable for shots earlier than 100775 (2000 JAN 17)')

    imas_version = getattr(ods, 'imas_version', None)
    separate_valves = imas_version is not None and compare_version(imas_version, "3.34.0") >= 0

    i = 0  # Pipe index
    iv = 0  # Valve index

    def pipe_copy(pipe_in):
        pipe_out = ods['gas_injection']['pipe'][i]
        for field in ['name', 'exit_position.r', 'exit_position.z', 'exit_position.phi']:
            pipe_out[field] = copy.copy(pipe_in[field])
        if separate_valves:
            valve_out = ods['gas_injection']['valve'][iv]
            pipe_out['valve_indices'] = [iv]
        else:
            valve_out = None
        return pipe_out, valve_out

    # PFX1
    pipe_indices = []
    for angle in [12, 139, 259]:  # degrees, DIII-D hardware left handed coords
        pipe_pfx1 = ods['gas_injection']['pipe'][i]
        pipe_pfx1['name'] = 'PFX1_{:03d}'.format(angle)
        pipe_pfx1['exit_position']['r'] = 1.286  # m
        pipe_pfx1['exit_position']['z'] = 1.279  # m
        pipe_pfx1['exit_position']['phi'] = -np.pi / 180.0 * angle  # radians, right handed
        if separate_valves:
            pipe_pfx1['valve_indices'] = [iv]
        else:
            pipe_pfx1['valve'][0]['identifier'] = 'PFX1'
        pipe_indices += [i]

        dr = -1.116 + 1.286
        dz = -1.38 + 1.279
        # pipea['exit_position']['direction'] = 180/np.pi * tan(dz/dr) if dr != 0 else 90 * sign(dz)
        pipe_pfx1['second_point']['phi'] = pipe_pfx1['exit_position']['phi']
        pipe_pfx1['second_point']['r'] = pipe_pfx1['exit_position']['r'] + dr
        pipe_pfx1['second_point']['z'] = pipe_pfx1['exit_position']['z'] + dz
        i += 1
    if separate_valves:
        valve_pfx1 = ods['gas_injection']['valve'][iv]
        valve_pfx1['identifier'] = 'PFX1'
        valve_pfx1['name'] = 'PFX1'
        valve_pfx1['pipe_indices'] = pipe_indices
    iv += 1

    # PFX2 injects at the same poloidal locations as PFX1, but at different toroidal angles
    pipe_indices = []
    for angle in [79, 199, 319]:  # degrees, DIII-D hardware left handed coords
        pipe_copy(pipe_pfx1)
        pipe_pfx2 = ods['gas_injection']['pipe'][i]
        pipe_pfx2['name'] = 'PFX2_{:03d}'.format(angle)
        pipe_pfx2['exit_position']['phi'] = -np.pi / 180.0 * angle  # rad
        if separate_valves:
            pipe_pfx2['valve_indices'] = [iv]
        else:
            pipe_pfx2['valve'][0]['identifier'] = 'PFX2'
        pipe_indices += [i]
        pipe_pfx2['second_point']['phi'] = pipe_pfx2['exit_position']['phi']
        i += 1
    if separate_valves:
        valve_pfx2 = ods['gas_injection']['valve'][iv]
        valve_pfx2['identifier'] = 'PFX2'
        valve_pfx2['name'] = 'PFX2'
        valve_pfx2['pipe_indices'] = pipe_indices
    iv += 1

    # GAS A
    pipea = ods['gas_injection']['pipe'][i]
    pipea['name'] = 'GASA_300'
    pipea['exit_position']['r'] = 1.941  # m
    pipea['exit_position']['z'] = 1.01  # m
    pipea['exit_position']['phi'] = -np.pi / 180.0 * 300  # rad
    # pipea['exit_position']['direction'] = 270.  # degrees, giving dir of pipe leading towards injector, up is 90
    pipea['second_point']['phi'] = pipea['exit_position']['phi']
    pipea['second_point']['r'] = pipea['exit_position']['r']
    pipea['second_point']['z'] = pipea['exit_position']['z'] - 0.01
    if separate_valves:
        pipea['valve_indices'] = [iv]
        valvea = ods['gas_injection']['valve'][iv]
        valvea['identifier'] = 'GASA'
        valvea['name'] = 'GASA'
        valvea['pipe_indices'] = [i]
    else:
        pipea['valve'][0]['identifier'] = 'GASA'
    i += 1
    iv += 1

    # GAS B injects in the same place as GAS A
    pipeb, valveb = pipe_copy(pipea)
    pipeb['name'] = 'GASB_300'
    if separate_valves:
        valveb['name'] = valveb['identifier'] = 'GASB'
        valveb['pipe_indices'] = [i]
    else:
        pipeb['valve'][0]['identifier'] = 'GASB'
    i += 1
    iv += 1

    # GAS C
    pipec = ods['gas_injection']['pipe'][i]
    pipec['name'] = 'GASC_000'
    pipec['exit_position']['r'] = 1.481  # m
    pipec['exit_position']['z'] = -1.33  # m
    pipec['exit_position']['phi'] = -np.pi / 180.0 * 0
    # pipec['exit_position']['direction'] = 90.  # degrees, giving direction of pipe leading towards injector
    pipec['second_point']['phi'] = pipec['exit_position']['phi']
    pipec['second_point']['r'] = pipec['exit_position']['r']
    pipec['second_point']['z'] = pipec['exit_position']['z'] + 0.01
    if separate_valves:
        valvec = ods['gas_injection']['valve'][iv]
        valvec['name'] = 'GASC'
        valvec['identifier'] = 'GASC'
        valvec['pipe_indices'] = [i]
        pipec['valve_indices'] = [iv]
    else:
        pipec['valve'][0]['identifier'] = 'GASC'

    i += 1
    iv += 1

    # GAS D injects at the same poloidal location as GAS A, but at a different toroidal angle.
    # There is a GASD piezo valve that splits into four injectors, all of which have their own gate valves and can be
    # turned on/off independently. Normally, only one would be used at at a time.
    pipe_copy(pipea)
    piped = ods['gas_injection']['pipe'][i]
    piped['name'] = 'GASD_225'  # This is the injector name
    piped['exit_position']['phi'] = -np.pi / 180.0 * 225
    piped['second_point']['phi'] = piped['exit_position']['phi']
    if separate_valves:
        valved = ods['gas_injection']['valve'][iv]
        valved['name'] = valved['identifier'] = 'GASD'  # This is the piezo name
        valved['pipe_indices'] = [i]
        piped['valve_indices'] = [iv]
    else:
        piped['valve'][0]['identifier'] = 'GASD'
    i += 1
    iv += 1

    # GASE is at the same poloidal location as GASC
    pipee, valvee = pipe_copy(pipec)
    if separate_valves:
        valvee['name'] = valvee['identifier'] = 'GASE'
        valvee['pipe_indices'] = [i]
    else:
        pipee['valve'][0]['identifier'] = 'GASE'
    i += 1
    iv += 1

    # Spare 225 is an extra branch of the GASD line after the GASD piezo (might not have a valve)
    pipe_copy(piped)
    pipes225 = ods['gas_injection']['pipe'][i]
    pipes225['name'] = 'Spare_225'  # This is the injector name
    i += 1

    # RF_170 and RF_190: gas ports near the 180 degree antenna, on the GASD line
    # I don't think these are connected anymore, but I don't know the shot number after which they were removed.
    for angle in [170, 190]:
        pipe_rf = ods['gas_injection']['pipe'][i]
        pipe_rf['name'] = 'RF_{:03d}'.format(angle)
        pipe_rf['exit_position']['r'] = 2.38  # m
        pipe_rf['exit_position']['z'] = -0.13  # m
        pipe_rf['exit_position']['phi'] = -np.pi / 180.0 * angle  # rad
        i += 1

    # DRDP
    pipe_copy(piped)
    pipedrdp = ods['gas_injection']['pipe'][i]
    pipedrdp['name'] = 'DRDP_225'
    if separate_valves:
        pipedrdp['valve_indices'] = [iv]
        valvedrdp = ods['gas_injection']['valve'][iv]
        valvedrdp['identifier'] = 'DRDP'
        valvedrdp['name'] = 'DRDP'
        valvedrdp['pipe_indices'] = [i]
    else:
        pipedrdp['valve'][0]['identifier'] = 'DRDP'
    i += 1
    iv += 1

    # UOB
    pipe_indices = []
    for angle in [45, 165, 285]:
        pipe_uob = ods['gas_injection']['pipe'][i]
        pipe_uob['name'] = 'UOB_{:03d}'.format(angle)
        pipe_uob['exit_position']['r'] = 1.517  # m
        pipe_uob['exit_position']['z'] = 1.267  # m
        pipe_uob['exit_position']['phi'] = -np.pi / 180.0 * angle
        pipe_indices += [i]
        if separate_valves:
            pipe_uob['valve_indices'] = [iv]
        else:
            pipe_uob['valve'][0]['identifier'] = 'UOB'
        # pipe_uob['exit_position']['direction'] = 270.  # degrees, giving dir of pipe leading to injector, up is 90
        i += 1
    if separate_valves:
        valve_uob = ods['gas_injection']['valve'][iv]
        valve_uob['name'] = valve_uob['identifier'] = 'UOB'
        valve_uob['pipe_indices'] = pipe_indices
    iv += 1

    # LOB1
    pipe_indices = []
    for angle in [30, 120]:
        pipe_lob1 = ods['gas_injection']['pipe'][i]
        pipe_lob1['name'] = 'LOB1_{:03d}'.format(angle)
        pipe_lob1['exit_position']['r'] = 1.941  # m
        pipe_lob1['exit_position']['z'] = -1.202  # m
        pipe_lob1['exit_position']['phi'] = -np.pi / 180.0 * angle
        if separate_valves:
            pipe_lob1['valve_indices'] = [iv]
        else:
            pipe_lob1['valve'][0]['identifier'] = 'LOB1'
        pipe_indices += [i]
        # pipe_lob1['exit_position']['direction'] = 180.  # degrees, giving dir of pipe leading to injector; up is 90
        i += 1
    if separate_valves:
        valve_lob1 = ods['gas_injection']['valve'][iv]
        valve_lob1['name'] = valve_lob1['identifier'] = 'LOB1'
        valve_lob1['pipe_indices'] = pipe_indices
    iv += 1

    # Spare 75 is an extra branch of the GASC line after the LOB1 piezo
    pipes75 = ods['gas_injection']['pipe'][i]
    pipes75['name'] = 'Spare_075'
    pipes75['exit_position']['r'] = 2.249  # m (approximate / estimated from still image)
    pipes75['exit_position']['z'] = -0.797  # m (approximate / estimated from still image)
    pipes75['exit_position']['phi'] = 75  # degrees, DIII-D hardware left handed coords
    if separate_valves:
        pipes75['valve_indices'] = pipe_lob1['valve_indices']
        valve_lob1['pipe_indices'] = np.append(valve_lob1['pipe_indices'], [i])
    else:
        pipes75['valve'][0]['identifier'] = 'LOB1'
    # pipes75['exit_position']['direction'] = 180.  # degrees, giving direction of pipe leading towards injector
    i += 1

    # RF_010 & 350
    for angle in [10, 350]:
        pipe_rf_lob1 = ods['gas_injection']['pipe'][i]
        pipe_rf_lob1['name'] = 'RF_{:03d}'.format(angle)
        pipe_rf_lob1['exit_position']['r'] = 2.38  # m
        pipe_rf_lob1['exit_position']['z'] = -0.13  # m
        pipe_rf_lob1['exit_position']['phi'] = -np.pi / 180.0 * angle
        if separate_valves:
            pipe_rf_lob1['valve_indices'] = pipe_lob1['valve_indices']
            valve_lob1['pipe_indices'] = np.append(valve_lob1['pipe_indices'], [i])
        else:
            pipe_rf_lob1['valve'][0]['identifier'] = 'LOB1'
        # pipe_rf10['exit_position']['direction'] = 180.  # degrees, giving dir of pipe leading to injector; up is 90
        i += 1

    # DiMES chimney
    pipe_dimesc = ods['gas_injection']['pipe'][i]
    pipe_dimesc['name'] = 'DiMES_Chimney_165'
    pipe_dimesc['exit_position']['r'] = 1.481  # m
    pipe_dimesc['exit_position']['z'] = -1.33  # m
    pipe_dimesc['exit_position']['phi'] = -np.pi / 180.0 * 165
    if separate_valves:
        pipe_dimesc['valve_indices'] = [iv]
        cpgas_valve = ods['gas_injection']['valve'][iv]
        cpgas_valve['identifier'] = '240R-2'
        cpgas_valve['name'] = '240R-2 (PCS use GASD)'
        cpgas_valve['pipe_indices'] = [i]
    else:
        pipe_dimesc['valve'][0]['identifier'] = '240R-2'
    # pipe_dimesc['exit_position']['direction'] = 90.  # degrees, giving dir of pipe leading towards injector, up is 90
    i += 1
    iv += 1

    # CPBOT
    pipe_cpbot = ods['gas_injection']['pipe'][i]
    pipe_cpbot['name'] = 'CPBOT_150'
    pipe_cpbot['exit_position']['r'] = 1.11  # m
    pipe_cpbot['exit_position']['z'] = -1.33  # m
    pipe_cpbot['exit_position']['phi'] = -np.pi / 180.0 * 150
    if separate_valves:
        pipe_cpbot['valve_indices'] = pipe_dimesc['valve_indices']
        cpgas_valve['pipe_indices'] = np.append(cpgas_valve['pipe_indices'], [i])
    else:
        pipe_cpbot['valve'][0]['identifier'] = '240R-2'
    # pipe_cpbot['exit_position']['direction'] = 0.  # degrees, giving dir of pipe leading towards injector, up is 90
    i += 1

    # LOB2 injects at the same poloidal locations as LOB1, but at different toroidal angles
    if separate_valves:
        valve_lob2 = ods['gas_injection']['valve'][iv]
        valve_lob2['identifier'] = 'LOB2'
        valve_lob2['name'] = 'LOB2'
        valve_lob2['pipe_indices'] = []
    for angle in [210, 300]:
        pipe_copy(pipe_lob1)
        pipe_lob2 = ods['gas_injection']['pipe'][i]
        pipe_lob2['name'] = 'LOB2_{:03d}'.format(angle)
        pipe_lob2['exit_position']['phi'] = -np.pi / 180.0 * angle  # degrees, DIII-D hardware left handed coords
        if separate_valves:
            pipe_lob2['valve_indices'] = [iv]
            valve_lob2['pipe_indices'] = np.append(valve_lob2['pipe_indices'], [i])
        else:
            pipe_lob2['valve'][0]['identifier'] = 'LOB2'
        i += 1
    iv += 1

    # Dimes floor tile 165
    pipe_copy(pipec)
    pipe_dimesf = ods['gas_injection']['pipe'][i]
    pipe_dimesf['name'] = 'DiMES_Tile_160'
    pipe_dimesf['exit_position']['phi'] = -np.pi / 180.0 * 165
    if separate_valves:
        pipe_dimesf['valve_indices'] = pipe_lob2['valve_indices']
        valve_lob2['pipe_indices'] = np.append(valve_lob2['pipe_indices'], [i])
    else:
        pipe_dimesf['valve'][0]['identifier'] = 'LOB2'
    i += 1

    # RF COMB
    pipe_rfcomb = ods['gas_injection']['pipe'][i]
    pipe_rfcomb['name'] = 'RF_COMB_'
    pipe_rfcomb['exit_position']['r'] = 2.38  # m
    pipe_rfcomb['exit_position']['z'] = -0.13  # m
    # pipe_rfcomb['exit_position']['phi'] = Unknown, sorry
    if separate_valves:
        pipe_rfcomb['valve_indices'] = pipe_lob2['valve_indices']
        valve_lob2['pipe_indices'] = np.append(valve_lob2['pipe_indices'], [i])
    else:
        pipe_rfcomb['valve'][0]['identifier'] = 'LOB2'
    # pipe_rf307['exit_position']['direction'] = 180.  # degrees, giving dir of pipe leading towards injector, up is 90
    i += 1

    # RF307
    pipe_rf307 = ods['gas_injection']['pipe'][i]
    pipe_rf307['name'] = 'RF_307'
    pipe_rf307['exit_position']['r'] = 2.38  # m
    pipe_rf307['exit_position']['z'] = -0.13  # m
    pipe_rf307['exit_position']['phi'] = -np.pi / 180.0 * 307
    if separate_valves:
        pipe_rf307['valve_indices'] = pipe_lob2['valve_indices']
        valve_lob2['pipe_indices'] = np.append(valve_lob2['pipe_indices'], [i])
    else:
        pipe_rf307['valve'][0]['identifier'] = 'LOB2'
    # pipe_rf307['exit_position']['direction'] = 180.  # degrees, giving dir of pipe leading towards injector, up is 90
    i += 1

    # GAS H injects in the same place as GAS C
    pipe_copy(pipec)
    pipeh = ods['gas_injection']['pipe'][i]
    pipeh['name'] = 'GASH_000'
    if separate_valves:
        pipeh['valve_indices'] = []  # This one's not on the manifold schematic
    i += 1

    # GAS I injects in the same place as GAS C
    pipe_copy(pipec)
    pipei = ods['gas_injection']['pipe'][i]
    pipei['name'] = 'GASI_000'
    if separate_valves:
        pipei['valve_indices'] = []  # This one's not on the manifold schematic
    i += 1

    # GAS J injects in the same place as GAS D
    pipe_copy(piped)
    pipej = ods['gas_injection']['pipe'][i]
    pipej['name'] = 'GASJ_225'
    if separate_valves:
        pipej['valve_indices'] = []  # This one's not on the manifold schematic
    i += 1

    # RF260
    pipe_rf260 = ods['gas_injection']['pipe'][i]
    pipe_rf260['name'] = 'RF_260'
    pipe_rf260['exit_position']['r'] = 2.38  # m
    pipe_rf260['exit_position']['z'] = 0.14  # m
    pipe_rf260['exit_position']['phi'] = -np.pi / 180.0 * 260
    if separate_valves:
        pipe_rf260['valve_indices'] = pipe_lob2['valve_indices']  # Seems to have been removed. May have been on LOB2, though.
        valve_lob2['pipe_indices'] = np.append(valve_lob2['pipe_indices'], [i])
    else:
        pipe_rf260['valve'][0]['identifier'] = 'LOB2'
    # pipe_rf260['exit_position']['direction'] = 180.  # degrees, giving dir of pipe leading towards injector, up is 90
    i += 1

    # CPMID
    pipe_cpmid = ods['gas_injection']['pipe'][i]
    pipe_cpmid['name'] = 'CPMID'
    pipe_cpmid['exit_position']['r'] = 0.9  # m
    pipe_cpmid['exit_position']['z'] = -0.2  # m
    if separate_valves:
        pipe_cpmid['valve_indices'] = []  # Seems to have been removed. Not on schematic.
    # pipe_cpmid['exit_position']['direction'] = 0.  # degrees, giving dir of pipe leading towards injector, up is 90
    i += 1

    return {}


# The magnetic hardware description function (for bpol probes and flux loops) has been generalized & is in omfit_omas.py

# The pulse_schedule hardware description function has been generalized and is in omfit_omas.py

# The position_control hardware description function (a subset of pulse_schedule) is in omfit_omas.py


@make_available
def setup_pf_active_hardware_description_d3d(ods, *args):
    r"""
    Adds DIII-D tokamak poloidal field coil hardware geometry to ODS
    :param ods: ODS instance

    :param \*args: catch unused args to allow a consistent call signature for hardware description functions

    :return: dict
        Information or instructions for follow up in central hardware description setup
    """
    from omfit_classes.omfit_omas_utils import pf_coils_to_ods

    # From  iris:/fusion/usc/src/idl/efitview/diagnoses/DIII-D/coils.dat , accessed 2018 June 08  D. Eldon
    fc_dat = np.array(
        [  # R        Z       dR      dZ    tilt1  tilt2
            [0.8608, 0.16830, 0.0508, 0.32106, 0.0, 0.0],  # 0 in the last column really means 90 degrees.
            [0.8614, 0.50810, 0.0508, 0.32106, 0.0, 0.0],
            [0.8628, 0.84910, 0.0508, 0.32106, 0.0, 0.0],
            [0.8611, 1.1899, 0.0508, 0.32106, 0.0, 0.0],
            [1.0041, 1.5169, 0.13920, 0.11940, 45.0, 0.0],
            [2.6124, 0.4376, 0.17320, 0.1946, 0.0, 92.40],
            [2.3733, 1.1171, 0.1880, 0.16920, 0.0, 108.06],
            [1.2518, 1.6019, 0.23490, 0.08510, 0.0, 0.0],
            [1.6890, 1.5874, 0.16940, 0.13310, 0.0, 0.0],
            [0.8608, -0.17370, 0.0508, 0.32106, 0.0, 0.0],
            [0.8607, -0.51350, 0.0508, 0.32106, 0.0, 0.0],
            [0.8611, -0.85430, 0.0508, 0.32106, 0.0, 0.0],
            [0.8630, -1.1957, 0.0508, 0.32106, 0.0, 0.0],
            [1.0025, -1.5169, 0.13920, 0.11940, -45.0, 0.0],
            [2.6124, -0.4376, 0.17320, 0.1946, 0.0, -92.40],
            [2.3834, -1.1171, 0.1880, 0.16920, 0.0, -108.06],
            [1.2524, -1.6027, 0.23490, 0.08510, 0.0, 0.0],
            [1.6889, -1.5780, 0.16940, 0.13310, 0.0, 0.0],
        ]
    )

    ods = pf_coils_to_ods(ods, fc_dat)

    for i in range(len(fc_dat[:, 0])):
        fcid = 'F{}{}'.format((i % 9) + 1, 'AB'[int(fc_dat[i, 1] < 0)])
        ods['pf_active.coil'][i]['name'] = ods['pf_active.coil'][i]['identifier'] = fcid
        ods['pf_active.coil'][i]['element.0.identifier'] = fcid

    return {}


@make_available
def setup_interferometer_hardware_description_d3d(ods, shot):
    """
    Writes DIII-D CO2 interferometer chord locations into ODS.

    The chord endpoints ARE NOT RIGHT. Only the R for vertical lines or Z for horizontal lines is right.

    Data sources:
    DIII-D webpage: https://diii-d.gat.com/diii-d/Mci accessed 2018 June 07  D. Eldon

    :param ods: an OMAS ODS instance

    :param shot: int

    :return: dict
        Information or instructions for follow up in central hardware description setup
    """

    # As of 2018 June 07, DIII-D has four interferometers
    # phi angles are compliant with odd COCOS
    ods['interferometer.channel.0.identifier'] = 'r0'
    r0 = ods['interferometer.channel.0.line_of_sight']
    r0['first_point.phi'] = r0['second_point.phi'] = 225 * (-np.pi / 180.0)
    r0['first_point.r'], r0['second_point.r'] = 3.0, 0.8  # These are not the real endpoints
    r0['first_point.z'] = r0['second_point.z'] = 0.0

    for i, r in enumerate([1.48, 1.94, 2.10]):
        ods['interferometer.channel'][i + 1]['identifier'] = 'v{}'.format(i + 1)
        los = ods['interferometer.channel'][i + 1]['line_of_sight']
        los['first_point.phi'] = los['second_point.phi'] = 240 * (-np.pi / 180.0)
        los['first_point.r'] = los['second_point.r'] = r
        los['first_point.z'], los['second_point.z'] = -1.8, 1.8  # These are not the real points

    for i in range(len(ods['interferometer.channel'])):
        ch = ods['interferometer.channel'][i]
        ch['line_of_sight.third_point'] = copy.copy(ch['line_of_sight.first_point'])

    if shot < 125406:
        printw(
            'DIII-D CO2 pointnames were different before shot 125406. The physical locations of the chords seems to '
            'have been the same, though, so there has not been a problem yet (I think).'
        )

    return {}


def find_thomson_lens_d3d(shot, hw_call_sys, hw_revision='blessed'):
    """Read the Thomson scattering hardware map to figure out which lens each chord looks through"""
    cal_call = '.ts.{}.header.calib_nums'.format(hw_revision)
    cal_set = OMFITmdsValue(server='DIII-D', treename='ELECTRONS', shot=shot, TDI=cal_call).data()[0]
    hwi_call = '.{}.hwmapints'.format(hw_call_sys)
    printq('  Reading hw map int values: treename = "tscal", cal_set = {}, hwi_call = {}'.format(cal_set, hwi_call))
    try:
        hw_ints = OMFITmdsValue(server='DIII-D', treename='tscal', shot=cal_set, TDI=hwi_call).data()
    except MDSplus.MdsException:
        printw('WARNING: Error reading Thomson scattering hardware map to determine which lenses were used!')
        return None
    else:
        if len(np.shape(hw_ints)) < 2:
            # Contingency needed for cases where all view-chords are taken off of divertor laser and reassigned to core
            hw_ints = hw_ints.reshape(1, -1)
        hw_lens = hw_ints[:, 2]
        return hw_lens


@make_available
def setup_thomson_scattering_hardware_description_d3d(ods, shot, revision='BLESSED'):
    """
    Gathers DIII-D Thomson measurement locations from MDSplus and loads them into OMAS

    :param revision: string
        Thomson scattering data revision, like 'BLESSED', 'REVISIONS.REVISION00', etc.

    :return: dict
        Information or instructions for follow up in central hardware description setup
    """
    printq('Setting up DIII-D Thomson locations...')

    tsdat = OMFITmds(server='DIII-D', treename='ELECTRONS', shot=shot)['TS'][revision]

    is_subsys = np.array([np.all([item in tsdat[k] for item in ['DENSITY', 'TEMP', 'R', 'Z']]) for k in list(tsdat.keys())])
    subsystems = np.array(list(tsdat.keys()))[is_subsys]

    i = 0
    for sub in subsystems:
        lenses = find_thomson_lens_d3d(shot, sub, revision)
        try:
            nc = len(tsdat[sub]['R'].data())
        except (TypeError, KeyError):
            nc = 0
        for j in range(nc):
            ch = ods['thomson_scattering']['channel'][i]
            ch['name'] = 'TS_{sub:}_r{lens:+0d}_{ch:}'.format(sub=sub.lower(), ch=j, lens=lenses[j] if lenses is not None else -9)
            ch['identifier'] = '{}{:02d}'.format(sub[0], j)
            for pos in ['R', 'Z', 'PHI']:
                ch['position'][pos.lower()] = tsdat[sub][pos].data()[j] * (-np.pi / 180.0 if pos == 'PHI' else 1)
            i += 1
    return {}


@make_available
def setup_charge_exchange_hardware_description_d3d(ods, shot, analysis_type='CERQUICK'):
    """
    Gathers DIII-D CER measurement locations from MDSplus and loads them into OMAS

    :param analysis_type: string
        CER analysis quality level like CERQUICK, CERAUTO, or CERFIT.  CERQUICK is probably fine.

    :return: dict
        Information or instructions for follow up in central hardware description setup
    """
    printq('Setting up DIII-D CER locations...')

    cerdat = OMFITmds(server='DIII-D', treename='IONS', shot=shot)['CER'][analysis_type]

    subsystems = np.array([k for k in list(cerdat.keys()) if 'CHANNEL01' in list(cerdat[k].keys())])

    i = 0
    for sub in subsystems:
        try:
            channels = [k for k in list(cerdat[sub].keys()) if 'CHANNEL' in k]
        except (TypeError, KeyError):
            channels = []
        for j, channel in enumerate(channels):
            inc = 0
            for pos in ['R', 'Z', 'VIEW_PHI']:
                postime = cerdat[sub][channel]['TIME'].data()
                posdat = cerdat[sub][channel][pos].data()
                if postime is not None:
                    inc = 1
                    ch = ods['charge_exchange']['channel'][i]
                    ch['name'] = 'imCERtang_{sub:}{ch:02d}'.format(sub=sub.lower()[0], ch=j + 1)
                    ch['identifier'] = '{}{:02d}'.format(sub[0], j + 1)
                    chpos = ch['position'][pos.lower().split('_')[-1]]
                    chpos['time'] = postime / 1000.0  # Convert ms to s
                    chpos['data'] = posdat * -np.pi / 180.0 if (pos == 'VIEW_PHI') and posdat is not None else posdat
            i += inc
    return {}


@make_available
def setup_bolometer_hardware_description_d3d(ods, shot):
    """
    Load DIII-D bolometer chord locations into the ODS

    Data sources:
    - iris:/fusion/usc/src/idl/efitview/diagnoses/DIII-D/bolometerpaths.pro
    - OMFIT-source/modules/_PCS_prad_control/SETTINGS/PHYSICS/reference/DIII-D/bolometer_geo , access 2018June11 Eldon

    :return: dict
        Information or instructions for follow up in central hardware description setup
    """
    printq('Setting up DIII-D bolometer locations...')

    if shot < 91000:
        xangle = (
            np.array(
                [
                    292.40,
                    288.35,
                    284.30,
                    280.25,
                    276.20,
                    272.15,
                    268.10,
                    264.87,
                    262.27,
                    259.67,
                    257.07,
                    254.47,
                    251.87,
                    249.27,
                    246.67,
                    243.81,
                    235.81,
                    227.81,
                    219.81,
                    211.81,
                    203.81,
                    195.81,
                    187.81,
                    179.80,
                    211.91,
                    206.41,
                    200.91,
                    195.41,
                    189.91,
                    184.41,
                    178.91,
                    173.41,
                    167.91,
                    162.41,
                    156.91,
                    156.30,
                    149.58,
                    142.86,
                    136.14,
                    129.77,
                    126.77,
                    123.77,
                    120.77,
                    117.77,
                    114.77,
                    111.77,
                    108.77,
                    102.25,
                ]
            )
            * np.pi
            / 180.0
        )  # Converted to rad

        xangle_width = None

        zxray = (
            np.array(
                [
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    124.968,
                    129.870,
                    129.870,
                    129.870,
                    129.870,
                    129.870,
                    129.870,
                    129.870,
                    129.870,
                    129.870,
                    -81.153,
                    -81.153,
                    -81.153,
                    -81.153,
                    -81.153,
                    -81.153,
                    -81.153,
                    -81.153,
                    -81.153,
                    -81.153,
                    -81.153,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                    -72.009,
                ]
            )
            / 100.0
        )  # Converted to m

        rxray = (
            np.array(
                [
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    196.771,
                    190.071,
                    190.071,
                    190.071,
                    190.071,
                    190.071,
                    190.071,
                    190.071,
                    190.071,
                    190.071,
                    230.720,
                    230.720,
                    230.720,
                    230.720,
                    230.720,
                    230.720,
                    230.720,
                    230.720,
                    230.720,
                    230.720,
                    230.720,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                    232.900,
                ]
            )
            / 100.0
        )  # Converted to m

    else:
        xangle = (
            np.array(
                [  # There is a bigger step before the very last channel. Found in two different sources.
                    269.4,
                    265.6,
                    261.9,
                    258.1,
                    254.4,
                    250.9,
                    247.9,
                    244.9,
                    241.9,
                    238.9,
                    235.9,
                    232.9,
                    228.0,
                    221.3,
                    214.5,
                    208.2,
                    201.1,
                    194.0,
                    187.7,
                    182.2,
                    176.7,
                    171.2,
                    165.7,
                    160.2,
                    213.7,
                    210.2,
                    206.7,
                    203.2,
                    199.7,
                    194.4,
                    187.4,
                    180.4,
                    173.4,
                    166.4,
                    159.4,
                    156.0,
                    149.2,
                    142.4,
                    135.8,
                    129.6,
                    126.6,
                    123.6,
                    120.6,
                    117.6,
                    114.6,
                    111.6,
                    108.6,
                    101.9,
                ]
            )
            * np.pi
            / 180.0
        )  # Converted to rad

        xangle_width = (
            np.array(
                [
                    3.082,
                    3.206,
                    3.317,
                    3.414,
                    3.495,
                    2.866,
                    2.901,
                    2.928,
                    2.947,
                    2.957,
                    2.960,
                    2.955,
                    6.497,
                    6.342,
                    6.103,
                    6.331,
                    6.697,
                    6.979,
                    5.510,
                    5.553,
                    5.546,
                    5.488,
                    5.380,
                    5.223,
                    3.281,
                    3.348,
                    3.402,
                    3.444,
                    3.473,
                    6.950,
                    6.911,
                    6.768,
                    6.526,
                    6.188,
                    5.757,
                    5.596,
                    5.978,
                    6.276,
                    6.490,
                    2.979,
                    2.993,
                    2.998,
                    2.995,
                    2.984,
                    2.965,
                    2.938,
                    2.902,
                    6.183,
                ]
            )
            * np.pi
            / 180.0
        )  # Angular full width of the view-chord: calculations assume it's a symmetric cone.

        zxray = (
            np.array(
                [
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    72.817,
                    82.332,
                    82.332,
                    82.332,
                    82.332,
                    82.332,
                    82.332,
                    82.332,
                    82.332,
                    82.332,
                    -77.254,
                    -77.254,
                    -77.254,
                    -77.254,
                    -77.254,
                    -77.254,
                    -77.254,
                    -77.254,
                    -77.254,
                    -77.254,
                    -77.254,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                    -66.881,
                ]
            )
            / 100.0
        )  # Converted to m

        rxray = (
            np.array(
                [
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    234.881,
                    231.206,
                    231.206,
                    231.206,
                    231.206,
                    231.206,
                    231.206,
                    231.206,
                    231.206,
                    231.206,
                    231.894,
                    231.894,
                    231.894,
                    231.894,
                    231.894,
                    231.894,
                    231.894,
                    231.894,
                    231.894,
                    231.894,
                    231.894,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                    234.932,
                ]
            )
            / 100.0
        )  # Converted to m

    line_len = 3  # m  Make this long enough to go past the box for all chords.

    phi = np.array([60, 75])[(zxray > 0).astype(int)] * -np.pi / 180.0  # Convert to CCW radians
    fan = np.array(['Lower', 'Upper'])[(zxray > 0).astype(int)]
    fan_offset = np.array([0, int(len(rxray) // 2)])[(zxray < 0).astype(int)].astype(int)

    for i in range(len(zxray)):
        cnum = i + 1 - fan_offset[i]
        ods['bolometer']['channel'][i]['identifier'] = '{}{:02d}'.format(fan[i][0], cnum)
        ods['bolometer']['channel'][i]['name'] = '{} fan ch#{:02d}'.format(fan[i], cnum)
        cls = ods['bolometer']['channel'][i]['line_of_sight']  # Shortcut
        cls['first_point.r'] = rxray[i]
        cls['first_point.z'] = zxray[i]
        cls['first_point.phi'] = phi[i]
        cls['second_point.r'] = rxray[i] + line_len * np.cos(xangle[i])
        cls['second_point.z'] = zxray[i] + line_len * np.sin(xangle[i])
        cls['second_point.phi'] = cls['first_point.phi']

    return {'postcommands': ['trim_bolometer_second_points_to_box(ods)']}


@make_available_if(compare_version(imas_version, '3.25.0') >= 0)  # Might actually be 3.24.X. Close enough for now.
def setup_langmuir_probes_hardware_description_d3d(ods, shot):
    """
    Load DIII-D Langmuir probe locations into an ODS

    :param ods: ODS instance

    :param shot: int

    :return: dict
        Information or instructions for follow up in central hardware description setup
    """
    import MDSplus

    if compare_version(ods.imas_version, '3.25.0') < 0:
        printe('langmuir_probes.embedded requires a newer version of IMAS. It was added by 3.25.0.')
        printe('ABORTED setup_langmuir_probes_hardware_description_d3d due to old IMAS version.')
        return {}

    ods['dataset_description.data_entry'].setdefault('machine', 'DIII-D')
    ods['dataset_description.data_entry'].setdefault('pulse', shot)

    tdi = r'GETNCI("\\langmuir::top.probe_*.r", "LENGTH")'
    # "LENGTH" is the size of the data, I think (in bits?). Single scalars seem to be length 12.
    printq('Setting up Langmuir probes hardware description, shot {}; checking availability, TDI={}'.format(shot, tdi))
    m = OMFITmdsValue(server='DIII-D', shot=shot, treename='LANGMUIR', TDI=tdi)
    try:
        data_present = m.data() > 0
    except MDSplus.MdsException:
        data_present = []
    nprobe = len(data_present)
    printq('Looks like up to {} Langmuir probes might have valid data for {}'.format(nprobe, shot))
    j = 0
    for i in range(nprobe):
        if data_present[i]:
            r = OMFITmdsValue(server='DIII-D', shot=shot, treename='langmuir', TDI=r'\langmuir::top.probe_{:03d}.r'.format(i))
            chk = r.check(debug=True, check_dim_of=-1)  # Don't check dimensions on these data
            if chk['result'] and r.data() > 0:
                # Don't bother gathering more if r is junk
                z = OMFITmdsValue(server='DIII-D', shot=shot, treename='langmuir', TDI=r'\langmuir::top.probe_{:03d}.z'.format(i))
                pnum = OMFITmdsValue(server='DIII-D', shot=shot, treename='langmuir', TDI=r'\langmuir::top.probe_{:03d}.pnum'.format(i))
                label = OMFITmdsValue(server='DIII-D', shot=shot, treename='langmuir', TDI=r'\langmuir::top.probe_{:03d}.label'.format(i))
                printq('  Probe i={i:}, j={j:}, label={label:} passed the check; r={r:}, z={z:}'.format(**locals()))
                ods['langmuir_probes.embedded'][j]['position.r'] = r.data()[0]
                ods['langmuir_probes.embedded'][j]['position.z'] = z.data()[0]
                ods['langmuir_probes.embedded'][j]['position.phi'] = np.NaN  # Didn't find this in MDSplus
                ods['langmuir_probes.embedded'][j]['identifier'] = 'PROBE_{:03d}: PNUM={}'.format(i, pnum.data()[0])
                ods['langmuir_probes.embedded'][j]['name'] = str(label.data()[0]).strip()
                j += 1
            else:
                printq('Probe i={i:}, j={j:}, r={r:} failed the check with chk={chk:}'.format(**locals()))
    return {}


@make_available
def find_active_d3d_probes(shot, allowed_probes=None):
    """
    Serves LP functions by identifying active probes (those that have actual data saved) for a given shot

    Sorry, I couldn't figure out how to do this with a server-side loop over all
    the probes, so we have to loop MDS calls on the client side. At least I
    resampled to speed that part up.

    This could be a lot faster if I could figure out how to make the GETNCI
    commands work on records of array signals.

    :param shot: int

    :param allowed_probes: int array
        Restrict the search to a certain range of probe numbers to speed things up
        These are the numbers of storage trees in MDSplus, not the physical probe numbers

    :return: list of ints
    """
    print(f'Searching for active DIII-D probes in shot {shot}...')
    omv_kw = dict(server='DIII-D', shot=shot, treename='langmuir')
    lp_paths_unfiltered = OMFITmdsValue(TDI=r'GETNCI("\\TOP.PROBE_*.TEMP", "MINPATH")', **omv_kw).data()
    lp_paths = [lpp.strip() for lpp in lp_paths_unfiltered if re.match('[0-9]', lpp.strip().split(':')[0].split('_')[-1])]
    print(f'{len(lp_paths)} possible probes found')
    lp_nums = np.array([int(lpp.strip().split(':')[0].split('_')[-1]) for lpp in lp_paths])
    idx = lp_nums.argsort()
    lp_nums = lp_nums[idx]
    lp_paths = np.array(lp_paths)[idx]
    if allowed_probes is None:
        acceptable_probe_numbers = lp_nums
        acceptable_paths = lp_paths
    else:
        acceptable_probe_numbers = np.array([lpn for lpn in lp_nums if lpn in allowed_probes])
        acceptable_paths = np.array([lp_paths[i] for i in range(len(lp_nums)) if lp_nums[i] in allowed_probes])
        print(f'{len(acceptable_paths)} candidates remain after limiting search to only some specific probes.')
    valid = [None] * len(acceptable_paths)
    OMFIT['acceptable_paths'] = acceptable_paths
    OMFIT['acceptable_probe_numbers'] = acceptable_probe_numbers
    OMFIT['lp_nums'] = lp_nums
    OMFIT['lp_paths'] = lp_paths

    i = wd = tc = 0
    for i in range(len(acceptable_paths)):
        ascii_progress_bar(
            i,
            0,
            len(acceptable_paths),
            mess=f'Checking DIII-D#{shot} LP probes for valid data {lp_paths[i]} valid={valid[i]}; '
            f'probes with valid data so far = {wd}/{tc}',
        )
        m = OMFITmdsValue(TDI=f'resample({acceptable_paths[i]}, 0, 1, 1)', **omv_kw)
        valid[i] = m.check() and np.any(m.data() > 0)
        wd = np.sum(np.array(valid).astype(bool))
        tc = len([a for a in valid if a is not None])
    ascii_progress_bar(
        i + 1,
        0,
        len(acceptable_paths),
        mess=f'Checked DIII-D#{shot} LP probes for valid data and found {wd} out of {tc} probes have valid data.',
    )
    print()
    return acceptable_probe_numbers[np.array(valid).astype(bool)]


@make_available_if(compare_version(imas_version, '3.25.0') >= 0)  # Might actually be 3.24.X. Close enough for now.
def load_data_langmuir_probes_d3d(
    ods, shot, probes=None, allowed_probes=None, tstart=0, tend=0, dt=0.0002, overwrite=False, quantities=None
):
    """
    Downloads LP probe data from MDSplus and loads them to the ODS

    :param ods: ODS instance

    :param shot: int

    :param probes: int array-like [optional]
        Integer array of DIII-D probe numbers.
        If not provided, find_active_d3d_probes() will be used.

    :param allowed_probes: int array-like [optional]
        Passed to find_active_d3d_probes(), if applicable.
        Improves speed by limiting search to a specific range of probe numbers.

    :param tstart: float
        Time to start resample (s)

    :param tend: float
        Time to end resample (s)
        Set to <= tstart to disable resample
        Server-side resampling does not work when time does not increase
        monotonically, which is a typical problem for DIII-D data. Resampling is
        not recommended for DIII-D.

    :param dt: float
        Resample interval (s)
        Set to 0 to disable resample
        Server-side resampling does not work when time does not increase
        monotonically, which is a typical problem for DIII-D data. Resampling is
        not recommended for DIII-D.

    :param overwrite: bool
        Download and write data even if they already are present in the ODS.

    :param quantities: list of strings [optional]
        List of quantities to gather. None to gather all available. Options are:
        ion_saturation_current, heat_flux_parallel, n_e, t_e', surface_area_effective,
        v_floating, and b_field_angle

    :return: ODS instance
        The data are added in-place, so catching the return is probably unnecessary.
    """

    tf = 1e3  # DIII-D stores LP data in ms, not s

    # Make sure the ODS is for the right device/shot
    device = 'DIII-D'
    ensure_consistent_experiment_info(ods, device, shot)

    # Make sure the ODS already has probe positions in it
    hw_ready = True
    try:
        if not is_numeric(ods['langmuir_probes.embedded[0].position.r']):
            hw_ready = False
        if 'name' not in ods['langmuir_probes.embedded.0']:
            hw_ready = False
    except Exception:
        hw_ready = False
    if not hw_ready:
        setup_langmuir_probes_hardware_description_d3d(ods, shot)

    # Select probes to gather
    probes = probes or find_active_d3d_probes(shot, allowed_probes=allowed_probes)

    do_resample = (dt > 0) and (tend > tstart)
    if do_resample:
        printe('DIII-D often has 0s @ end of timebase for LP data in MDSplus, breaking server-side resampling.')
        printw('Recommended: abort and set tend=tstart or dt=0 to disable resampling.')

    available_quantities = dict(
        ion_saturation_current='ISAT',
        heat_flux_parallel='HEATFLUX',
        n_e='DENS',
        t_e='TEMP',
        surface_area_effective='AREA',
        v_floating='POT',  # or is it 'v_plasma'?
        b_field_angle='ANGLE',
    )
    if quantities is None:
        quantities = available_quantities
    else:
        quantities = {k: v for k, v in available_quantities.items() if k in quantities}
    factors = dict(
        ion_saturation_current=1,  # A to A
        heat_flux_parallel=1e7,  # kW cm^-2 to W m^-2
        n_e=1e6,  # cm^-3 to m^-3
        t_e=1,  # eV to eV
        surface_area_effective=1e-4,  # cm^2 to m^2
        v_floating=1,  # V to V
        v_plasma=1,  # V to V
        b_field_angle=np.pi / 180.0,  # deg to rad
    )
    has_err = ['t_e', 'n_e', 'v_floating']
    i = 0
    for i, probe in enumerate(probes):
        p_idx = probe - 1
        tdi_base = rf'\TOP.PROBE_{probe:03d}'

        if not overwrite:
            try:
                _ = ods['langmuir_probes.embedded'][i]['time']
                _ = ods['langmuir_probes.embedded'][i]['ion_saturation_current.data']
            except Exception:
                pass
            else:
                printd(f'Skipping probe {probe} with index {p_idx} because it already has data', topic='omfit_omas_d3d')
                continue

        for j, ods_loc in enumerate(quantities):
            quantity = quantities[ods_loc]
            tdi = f'{tdi_base}.{quantity}'
            ascii_progress_bar(
                i * len(quantities) + j, 0, len(probes) * len(quantities), mess=f'Loading {device}#{shot} LP probe data ({probe}, {tdi})'
            )
            f = factors.get(ods_loc, 1.0)
            if ods_loc in has_err:
                tdi_err = f'{tdi}_ERR'
            else:
                tdi_err = None
            if do_resample:
                tdi = f'resample({tdi}, {tstart*tf}, {tend*tf}, {dt*tf})'
                if tdi_err is not None:
                    tdi_err = f'resample({tdi_err}, {tstart*tf}, {tend*tf}, {dt*tf})'

            m = OMFITmdsValue(server=device, shot=shot, treename='LANGMUIR', TDI=tdi)
            if tdi_err is not None:
                m_err = OMFITmdsValue(server=device, shot=shot, treename='LANGMUIR', TDI=tdi_err)
            else:
                m_err = None

            if m.check():
                ods['langmuir_probes.embedded'][i]['time'] = m.dim_of(0) / tf
                if m_err is not None:
                    ods['langmuir_probes.embedded'][i][ods_loc]['data'] = uarray(m.data() * f, abs(m_err.data()) * f)
                else:
                    ods['langmuir_probes.embedded'][i][ods_loc]['data'] = m.data() * f  # Just nominal values
            else:
                printd(
                    f'Probe {i} (#{p_idx:03d}) does not appear to have good {quantity} ({ods_loc}) data. Skipping...',
                    topic='omfit_omas_d3d',
                )
                printd(f'    TDI = {tdi}', topic='omfit_omas_d3d')

    ascii_progress_bar(i + 1, 0, len(probes), mess=f'Loaded {device}#{shot} LP probe data')
    if do_resample:
        printe('DIII-D often has 0s @ end of timebase for LP data in MDSplus, breaking server-side resampling.')
        printw('Recommended: check your data and repeat with dt=0 or tend=tstart if you find problems.')

    return ods


class OMFITd3dcompfile(SortedDict, OMFITascii):
    """
    OMFIT class to read DIII-D compensation files such as btcomp.dat ccomp.dat and icomp.dat
    """

    def __init__(self, filename, use_leading_comma=None, **kw):
        r"""
        OMFIT class to parse DIII-D MHD device files

        :param filename: filename

        :param \**kw: arguments passed to __init__ of OMFITascii
        """
        OMFITascii.__init__(self, filename, **kw)
        SortedDict.__init__(self)
        self.dynaLoad = True

    @dynaLoad
    def load(self):
        self.clear()
        with open(self.filename, 'r') as f:
            lines = f.readlines()

        for line in lines:
            linesplit = line.split()

            try:
                compshot = int(eval(linesplit[0]))
                self[compshot] = {}
                for compsig in linesplit[1:]:
                    self[compshot][compsig.strip("'")] = {}

            except Exception:
                sig = linesplit[0][1:].strip()
                comps = [float(i) for i in linesplit[2:]]
                for comp, compsig in zip(comps, self[compshot]):
                    self[compshot][compsig][sig] = comp

        return self



try:
    # framework is running
    from .startup_choice import *
except ImportError as _excp:
    # class is imported by itself
    if (
        'attempted relative import with no known parent package' in str(_excp)
        or 'No module named \'omfit_classes\'' in str(_excp)
        or "No module named '__main__.startup_choice'" in str(_excp)
    ):
        from startup_choice import *
    else:
        raise

from omfit_classes.omfit_path import OMFITpath

import xarray as xr
import pandas as pd

__all__ = ['OMFITmatrix']


class OMFITmatrix(SortedDict, OMFITpath):
    """
    OMFITmatrix leverages both xarray and pandas as an efficient way of storing matrices to file.
    Internally, the data is stored as an xarray.DataArray under self['data'].
    """

    def __init__(self, filename, bin=None, zip=None, **kw):
        r"""
        :param filename: path to file.

        :param bin: def None, filetype is unknown,
                    if True,              NetCDF,
                    if False,             ASCII.

        :param zip: def None, compression is unknown,
                    if False,                switched off,
                    if True,                           on.

        :param \**kw: keywords for OMFITpath.__init__()
        """

        SortedDict.__init__(self)
        OMFITpath.__init__(self, filename, **kw)
        self.OMFITproperties['bin'] = bin
        self.OMFITproperties['zip'] = zip
        self.dynaLoad = True

    def __tree_repr__(self):

        r = 'FILE: %s   (%s' % (os.path.split(self.filename)[1], sizeof_fmt(self.filename, ' '))
        try:
            assert self.OMFITproperties['bin'] is not None
            if self.OMFITproperties['zip']:
                r += ', ' + ('zlibbed' if self.OMFITproperties['bin'] else 'gzipped')
            else:
                r += ', plain'
            r += ' ' + ('NetCDF' if self.OMFITproperties['bin'] else 'ASCII')
        except Exception:
            pass
        r += ')'
        try:
            assert np.shape(self['data'])
            r += '   DATA: ' + 'x'.join(['{:,}'.format(s) for s in np.shape(self['data'])])
        except Exception:
            pass
        return r, []

    @dynaLoad
    def load(self, bin=None, zip=None, xrkw={}, **pdkw):
        r"""
        https://xarray.pydata.org/en/stable/generated/xarray.open_dataarray.html
        https://pandas.pydata.org/pandas-docs/stable/generated/pandas.read_csv.html

        :param bin: def None, load through xarray first, then through pandas,
                    if True,               xarray only,
                    if False,              pandas only.

        :param zip: def False, compression is switched off,
                    if True,                           on.

        :param    xrkw: keywords for xarray.open_dataarray()

        :param \**pdkw: keywords for pandas.read_csv()

        :return bin, zip: resulting values for binary, zipped.
        """

        # allow creation of empty files
        if not os.path.exists(self.filename) or not os.path.getsize(self.filename):
            self['data'] = None
            return None, None

        # load defaults
        if bin is None:
            bin = self.OMFITproperties['bin']
        if zip is None:
            zip = self.OMFITproperties['zip']

        # read NetCDF through xarray
        try:

            if bin is False:
                raise Exception('binary only')

            xrkw.setdefault('engine', 'netcdf4')
            self['data'] = xr.open_dataarray(self.filename, **xrkw)
            self['data'].values
            self['data'].close()

            bin = True
            try:
                zip = self['data'].encoding['zlib']
            except Exception:
                zip = False

        # read ASCII through pandas
        except Exception as e:

            if bin is True or not np.any([m in repr(e) for m in ['binary only', 'NetCDF: Unknown file format']]):
                raise

            pdkw.setdefault('header', None)
            pdkw.setdefault('delim_whitespace', True)
            pdkw.setdefault('index_col', False)
            pdkw.setdefault('compression', 'gzip' if zip else None)

            self['data'] = xr.DataArray(pd.read_csv(self.filename, **pdkw))

            bin = False
            zip = pdkw['compression'] == 'gzip'

        finally:

            # save defaults
            self.OMFITproperties['bin'] = bin
            self.OMFITproperties['zip'] = zip

        return bin, zip

    @dynaSave
    def save(self, bin=None, zip=None, xrkw={}, **pdkw):
        r"""
        https://xarray.pydata.org/en/stable/generated/xarray.DataArray.to_netcdf.html
        https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.to_csv.html

        :param bin: def None, save through xarray first, then pandas,
                    if True,               xarray only,
                    if False,              pandas only.

        :param zip: def False, compression is switched off,
                    if True,                           on.

        :param    xrkw: keywords for xarray.to_netcdf()

        :param \**pdkw: keywords for pandas.to_csv()

        :return bin, zip: resulting values for binary, zipped.
        """

        if self.dynaLoad:
            return None, None

        # load defaults
        if bin is None:
            bin = self.OMFITproperties['bin']
        if zip is None:
            zip = self.OMFITproperties['zip']

        # save NetCDF through xarray
        try:

            if bin is False:
                raise Exception('binary only')

            self['data'].encoding = {'zlib': bool(zip), 'shuffle': bool(zip), 'complevel': int(bool(zip))}

            self['data'].to_netcdf(self.filename, **xrkw)
            self['data'].close()

            bin = True

        # save ASCII through pandas
        except Exception as e:

            if bin is True or str(e) != 'binary only':
                raise

            pdkw.setdefault('header', None)
            pdkw.setdefault('sep', ' ')
            pdkw.setdefault('index', False)
            pdkw.setdefault('compression', 'gzip' if zip else None)
            pdkw.setdefault('na_rep', 'nan')

            self['data'].to_pandas().to_csv(self.filename, **pdkw)

            bin = False

        finally:

            # save defaults
            self.OMFITproperties['bin'] = bin
            self.OMFITproperties['zip'] = bool(zip)

        return bin, zip

    def close(self, **kw):
        r"""
        :param \**kw: keywords for self.save()
        """

        if self.dynaLoad:
            return
        self.save(**kw)
        SortedDict.clear(self)
        self.dynaLoad = True



import sys
from omfit_classes import unix_os as os
import argparse

try:
    OMFITstartDir
except Exception:
    OMFITstartDir = os.getcwd()


def omfit_parse_args(input_list=None, show_omfit_launch_options=False):
    """
    Parse the arguments in input_list

    :param input_list: A list of strings, such as sys.argv, to be parsed for arguments passed to OMFIT

    :param show_omfit_launch_options: Display options that are parsed by omfit_launch.sh script or not

    :returns: The results of argparse.ArgumentParser(input_list).parse_args() with smart modifications to get paths correct
    """

    if input_list is None:
        input_list = sys.argv[1:]

    parser = argparse.ArgumentParser(description='OMFIT (One Modeling Framework for Integrated Tasks)')

    parser.add_argument('-s', '--setup', action='store_true', help='Helper setup for SSH authentication and dependencies')
    parser.add_argument('--purge', action='store_true', help='Purge all files in temporary OMFIT locations')
    parser.add_argument('--reset', action='store_true', help='Reset OMFIT SSH/MDS/SQL connections')
    parser.add_argument('--packages', action='store_true', help='List versions of OMFIT required packages')
    if show_omfit_launch_options:
        parser.add_argument('--cwd', action='store_true', help='use current working directory as $OMFIT_ROOT')
        parser.add_argument('-P', '--Python', action='store_true', help='Python major/minor release of executable to use (python[3, 3.x])')
    parser.add_argument('-g', '--scriptGui', action='store_true', help='Run the specified script in the GUI')
    parser.add_argument('-p', '--project', action='store', help='Open the specified project (`-1` for most recent project)')
    parser.add_argument('-m', '--module', action='store', nargs='*', help='Load the specified module(s)', metavar='REMOTE/BRANCH:MODULE')
    parser.add_argument(
        '-M', '--Module', action='store', nargs='*', help='Load the specified module(s) in developer mode', metavar='REMOTE/BRANCH:MODULE'
    )
    parser.add_argument('-c', '--command', action='store', nargs='*', help='Execute the specified Python command(s)')
    parser.add_argument('-b', '--bare', action='store_true', help='Do not start framework')
    parser.add_argument('scriptFile', nargs='?', help='Run the specified script (optional)', default=None)
    parser.add_argument('scriptArgs', nargs='*', help='Arguments passed to script (optional)')

    args = parser.parse_args(input_list)

    OMFITsrc = os.path.split(__file__)[0]

    # pre-process project path
    if args.project:

        if args.project.strip() != '-1':

            # prepend starting directory
            if args.project[0] != '/':
                args.project = os.path.sep.join([OMFITstartDir, args.project])

            # extract basename
            basename = os.path.split(args.project)[1]

            # check existence
            if not os.path.exists(args.project):
                raise IOError('Could not open project: %s.' % args.project)

            # update absolute path
            args.project = os.path.abspath(args.project)

    # pre-process script path
    if args.scriptFile:

        # prepend starting directory
        if args.scriptFile[0] != '/':
            args.scriptFile = os.path.sep.join([OMFITstartDir, args.scriptFile])

        # extract basename
        basename = os.path.split(args.scriptFile)[1]

        # check existence
        if not os.path.exists(args.scriptFile):
            if os.path.exists(os.sep.join([OMFITsrc, '..', 'scripts', os.path.splitext(basename)[0] + '.py'])):
                args.scriptFile = os.sep.join([OMFITsrc, '..', 'scripts', os.path.splitext(basename)[0] + '.py'])
            elif os.path.exists(os.sep.join([OMFITsrc, '..', 'regression', os.path.splitext(basename)[0] + '.py'])):
                args.scriptFile = os.sep.join([OMFITsrc, '..', 'regression', os.path.splitext(basename)[0] + '.py'])
            else:
                raise IOError('Could not open script: %s.' % args.scriptFile)

        # update absolute path
        args.scriptFile = os.path.abspath(args.scriptFile)

    return args


def nice_script_args(args):
    """
    Process script arguments
    """
    import shlex

    s = shlex.shlex(' '.join(args.scriptArgs), posix=False)
    s.whitespace_split = False
    s = list(s) + [None]

    kwargs = {}
    vaname = None
    for k, val in enumerate(s[:-1]):
        if val == '=':
            pass
        elif k == 0 or s[k + 1] == '=':
            varname = val
        else:
            if varname not in kwargs:
                kwargs[varname] = val
            else:
                kwargs[varname] = kwargs[varname] + val

    for varname in kwargs:
        try:
            kwargs[varname] = eval(kwargs[varname])
        except Exception:
            raise Exception(f'Problem with parsing command line argument: {varname}={kwargs[varname]}')
    return kwargs



from omfit_classes import unix_os as os
import sys
import platform
import subprocess
import distutils.spawn
import re
import time

import tkinter as tk
from tkinter import ttk
import tkinter.font as tkFont


def b2s(bytes):
    return bytes.decode("utf-8")


os.chdir(os.environ['HOME'])


def OMFITfont(ftype, size=0, family=''):
    """
    :param ftype: 'normal' or 'bold'

    :param size: positive or negative number relative to the tkinter default

    :param family: family font

    :return: tk font object to be used in tkinter calls
    """
    tmp = ttk.Label(text="bla")
    f = tkFont.Font(font=tmp['font'])
    tmp.destroy()
    f['size'] = -(abs(f['size']) + size)
    if ftype == 'normal':
        pass
    else:
        f['weight'] = ftype
    if family:
        f['family'] = family
    return f


def is_running(process):
    """
    This function returns True or False depending on whether a process is running or not

    This relies on grep of the `ps axw` command

    :param process: string with process name or process ID

    :return: False if process is not running, otherwise line of `ps` command
    """
    process = str(process)
    s = subprocess.Popen(
        [distutils.spawn.find_executable('ps'), "axw"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    std_out, std_err = list(map(b2s, s.communicate()))
    for item in std_out.split('\n'):
        if re.search(' ' + process + ' ', item) or re.search(r'^' + process + ' ', item):
            return item
    return False


def sayGoodbye(mainPid):
    try:
        os.kill(mainPid, 2)
    except Exception:
        pass
    sys.exit()


mainPid = None
try:
    mainPid = int(sys.argv[1])
except Exception:
    mainPid = quit()

try:
    session_color = re.sub('\\', '', sys.argv[2])
except Exception:
    session_color = None

root = tk.Tk()
root.withdraw()
root.wm_minsize(400, 10)
root.wm_resizable(0, 0)
root.wm_title('Halt OMFIT PID ' + str(mainPid) + ' on ' + platform.uname()[1].split('.')[0])
root.protocol("WM_DELETE_WINDOW", 'break')

allertText = tk.StringVar()
allertText.set('')
ttk.Label(root, textvariable=allertText, justify=tk.LEFT, font=OMFITfont('normal', 0, 'Courier')).pack(
    side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=2, pady=2
)
if session_color is not None:
    tk.Frame(root, background=session_color, height=2).pack(side=tk.TOP, expand=tk.YES, fill=tk.X, padx=2, pady=2)
ttk.Button(root, text='Interrupt OMFIT instance with PID ' + str(mainPid), command=lambda: sayGoodbye(mainPid)).pack(
    side=tk.TOP, expand=tk.YES, fill=tk.X, padx=2, pady=2
)


def stats(mainPid):
    if platform.system() == 'Darwin':
        topstr = distutils.spawn.find_executable('top') + ' -l 1 -ncols 4 -o cpu -o time -pid ' + str(mainPid)
    else:
        topstr = distutils.spawn.find_executable('top') + ' -b -n 1 -p ' + str(mainPid)
    s = subprocess.Popen(topstr, shell=True, stdout=subprocess.PIPE)
    allertText.set(s.stdout.read())
    if is_running(mainPid):
        root.after(1000, lambda: stats(mainPid))
    else:
        root.destroy()
        sys.exit()


stats(mainPid)

if platform.system() == 'Darwin':
    root.deiconify()
    root.lower()
else:
    root.iconify()

root.mainloop()



# -*-Python-*-
# Created by izzov at 24 Feb 2021  05:48

"""
This GUI is used to make tweaks to the current axes after the figure is created
in order to spruce it up for presenting or publishing
"""

OMFITx.Label('All changes apply to the current axes')

# global


def add_grid(location):
    grid(scratch['grid'])


def set_aspect(location):
    axis(scratch['aspect'])


def entitle(location):
    ax = gca()
    scratch.setdefault('tfs', 14)
    ax.set_title(scratch['title'], fontsize=scratch['tfs'])


with OMFITx.same_row():
    ax = gca()
    if ax.get_title():
        scratch['title'] = ax.get_title()
    OMFITx.Entry("scratch['title']", lbl='title', default="Plasma Current", postcommand=entitle)
    OMFITx.Entry("scratch['tfs']", lbl='font size', default=14, postcommand=entitle)

with OMFITx.same_row():
    OMFITx.ComboBox("scratch['grid']", {'on': 'on', 'off': False}, lbl='grid', default='off', postcommand=add_grid)
    OMFITx.ComboBox(
        "scratch['aspect']", {'equal': 'equal', 'image': 'image', 'auto': 'auto'}, lbl='aspect', default='auto', postcommand=set_aspect
    )


# x-axis


def set_xlim(location):
    ax = gca()
    ax.set_xlim(scratch['xlim'])


def set_xscale(location):
    ax = gca()
    ax.set_xscale(scratch['xscale'])


def set_xticks(location):
    ax = gca()
    ax.set_xticks(scratch['xticks'])


def set_xticklabels(location):
    ax = gca()
    scratch.setdefault('xfs', 12)
    ax.set_xticklabels(scratch['xtick_labels'], fontsize=scratch['xfs'])


def set_xlabel(location):
    ax = gca()
    scratch.setdefault('xfs', 12)
    ax.set_xlabel(scratch['xlabel'], fontsize=scratch['xfs'])


OMFITx.Tab('x-axis')

with OMFITx.same_row():
    ax = gca()
    if ax.get_xlabel():
        scratch['xlabel'] = ax.get_xlabel()
    OMFITx.Entry("scratch['xlabel']", lbl='xlabel', default="$\mathregular{t-t_{0}}$ $[{\mu}s]$", postcommand=set_xlabel)
    OMFITx.Entry("scratch['xfs']", lbl='font size', default=12, postcommand=set_xlabel)

with OMFITx.same_row():
    ax = gca()
    scratch['xlim'] = ax.get_xlim()
    OMFITx.Entry("scratch['xlim']", lbl='xlim', default=[0, 1], postcommand=set_xlim)
    OMFITx.ComboBox("scratch['xscale']", {'linear': 'linear', 'log': 'log'}, lbl='xscale', default='linear', postcommand=set_xscale)
with OMFITx.same_row():
    ax = gca()
    if any(ax.get_xticks()):
        scratch['xticks'] = ax.get_xticks()
    if any(ax.get_xticklabels()):
        scratch['xtick_labels'] = ax.get_xticklabels()
    elif any(ax.get_xticks()):
        scratch['xtick_labels'] = ax.get_xticks()
    OMFITx.Entry("scratch['xticks']", lbl='xticks', default=linspace(0, 1, 5), postcommand=set_xticks)
    OMFITx.Entry("scratch['xtick_labels']", lbl='xtick labels', default=linspace(0, 1, 5), postcommand=set_xticklabels)

# y-axis


def set_ylim(location):
    ax = gca()
    ax.set_ylim(scratch['ylim'])


def set_yscale(location):
    ax = gca()
    ax.set_yscale(scratch['yscale'])


def set_yticks(location):
    ax = gca()
    ax.set_yticks(scratch['yticks'])


def set_yticklabels(location):
    ax = gca()
    scratch.setdefault('yfs', 12)
    ax.set_yticklabels(scratch['ytick_labels'], fontsize=scratch['yfs'])


def set_ylabel(location):
    ax = gca()
    scratch.setdefault('yfs', 12)
    ax.set_ylabel(scratch['ylabel'], fontsize=scratch['yfs'])


OMFITx.Tab('y-axis')

with OMFITx.same_row():
    ax = gca()
    if ax.get_ylabel():
        scratch['ylabel'] = ax.get_ylabel()
    OMFITx.Entry("scratch['ylabel']", lbl='ylabel', default="$\mathregular{I_{p}}$ [MA]", postcommand=set_ylabel)
    OMFITx.Entry("scratch['yfs']", lbl='font size', default=12, postcommand=set_ylabel)

with OMFITx.same_row():
    ax = gca()
    scratch['ylim'] = ax.get_ylim()
    OMFITx.Entry("scratch['ylim']", lbl='ylim', default=[0, 1], postcommand=set_ylim)
    OMFITx.ComboBox("scratch['yscale']", {'linear': 'linear', 'log': 'log'}, lbl='yscale', default='linear', postcommand=set_yscale)
with OMFITx.same_row():
    ax = gca()
    if any(ax.get_yticks()):
        scratch['yticks'] = ax.get_yticks()
    if any(ax.get_yticklabels()):
        scratch['ytick_labels'] = ax.get_yticklabels()
    elif any(ax.get_yticks()):
        scratch['ytick_labels'] = ax.get_yticks()
    OMFITx.Entry("scratch['yticks']", lbl='yticks', default=linspace(0, 1, 5), postcommand=set_yticks)
    OMFITx.Entry("scratch['ytick_labels']", lbl='ytick labels', default=linspace(0, 1, 5), postcommand=set_yticklabels)


# add vertical and horizontal lines


def add_vertical_lines(location):
    ax = gca()
    yl = ax.get_ylim()
    scratch['vls'] = []
    scratch.setdefault('lstyle', '--')
    scratch.setdefault('lcolor', 'k')
    for xvl in scratch['xvls']:
        scratch['vls'].append(ax.plot([xvl, xvl], yl, linestyle=scratch['lstyle'], color=scratch['lcolor']))


def remove_vertical_lines():
    ax = gca()
    if scratch['vls']:
        for vl in scratch['vls']:
            line = vl.pop(0)
            line.remove()


OMFITx.Tab('Add vertical lines')

with OMFITx.same_row():
    OMFITx.Entry("scratch['lcolor']", lbl='color', default='k')
    OMFITx.Entry("scratch['lstyle']", lbl='style', default='--')
with OMFITx.same_row():
    OMFITx.Entry("scratch['xvls']", lbl='add at x', default=[0.5, 1.0], postcommand=add_vertical_lines)
    OMFITx.Button('remove all', remove_vertical_lines)


def add_horizontal_lines(location):
    ax = gca()
    xl = ax.get_xlim()
    scratch['hls'] = []
    scratch.setdefault('lstyle', '--')
    scratch.setdefault('lcolor', 'k')
    for yhl in scratch['yhls']:
        scratch['hls'].append(ax.plot(xl, [yhl, yhl], linestyle=scratch['lstyle'], color=scratch['lcolor']))


def remove_horizontal_lines():
    ax = gca()
    if scratch['hls']:
        for hl in scratch['hls']:
            line = hl.pop(0)
            line.remove()


OMFITx.Tab('Add horizontal lines')

with OMFITx.same_row():
    OMFITx.Entry("scratch['lcolor']", lbl='color', default='k')
    OMFITx.Entry("scratch['lstyle']", lbl='style', default='--')
with OMFITx.same_row():
    OMFITx.Entry("scratch['yhls']", lbl='add at y', default=[0.5, 1.0], postcommand=add_horizontal_lines)
    OMFITx.Button('remove all', remove_horizontal_lines)
